#  SPDX-FileCopyrightText: 2022 easyCore contributors  <core@easyscience.software>
#  SPDX-License-Identifier: BSD-3-Clause
#  © 2021-2022 Contributors to the easyCore project <https://github.com/easyScience/easyCore>

from __future__ import annotations

__author__ = "github.com/wardsimon"
__version__ = "0.1.0"

from abc import abstractmethod, ABCMeta

import weakref
from asteval import Interpreter
from numbers import Number
from typing import List, Union, Callable, TYPE_CHECKING, Optional, TypeVar

from easyCore import borg, np
from easyCore.Utils.json import MSONable

if TYPE_CHECKING:
    from easyCore.Objects.Variable import V


class ConstraintBase(MSONable, metaclass=ABCMeta):
    """
    A base class used to describe a constraint to be applied to easyCore base objects.
    """

    _borg = borg

    def __init__(
        self,
        dependent_obj: V,
        independent_obj: Optional[Union[V, List[V]]] = None,
        operator: Optional[Union[str, List[str]]] = None,
        value: Optional[Number] = None,
    ):
        self.aeval = Interpreter()
        self.dependent_obj_ids = self.get_key(dependent_obj)
        self.independent_obj_ids = None
        self._enabled = True
        self.external = False
        self._finalizer = None
        if independent_obj is not None:
            if isinstance(independent_obj, list):
                self.independent_obj_ids = [
                    self.get_key(obj) for obj in independent_obj
                ]
                if self.dependent_obj_ids in self.independent_obj_ids:
                    raise AttributeError(
                        "A dependent object can not be an independent object"
                    )
            else:
                self.independent_obj_ids = self.get_key(independent_obj)
                if self.dependent_obj_ids == self.independent_obj_ids:
                    raise AttributeError(
                        "A dependent object can not be an independent object"
                    )
            # Test if dependent is a parameter or a descriptor.
            # We can not import `Parameter`, so......
            if dependent_obj.__class__.__name__ == "Parameter":
                if not dependent_obj.enabled:
                    raise AssertionError(
                        "A dependent object needs to be initially enabled."
                    )
                if borg.debug:
                    print(
                        f"Dependent variable {dependent_obj}. It should be a `Descriptor`."
                        f"Setting to fixed"
                    )
                dependent_obj.enabled = False
                self._finalizer = weakref.finalize(
                    self, cleanup_constraint, self.dependent_obj_ids, True
                )

        self.operator = operator
        self.value = value

    @property
    def enabled(self) -> bool:
        """
        Is the current constraint enabled.

        :return: Logical answer to if the constraint is enabled.
        """
        return self._enabled

    @enabled.setter
    def enabled(self, enabled_value: bool):
        """
                Set the enabled state of the constraint. If the new value is the same as the current value only the state is
                changed.

        ... note:: If the new value is ``True`` the constraint is also applied after enabling.

                :param enabled_value: New state of the constraint.
                :return: None
        """

        if self._enabled == enabled_value:
            return
        elif enabled_value:
            self.get_obj(self.dependent_obj_ids).enabled = False
            self()
        else:
            self.get_obj(self.dependent_obj_ids).enabled = True
        self._enabled = enabled_value

    def __call__(self, *args, no_set: bool = False, **kwargs):
        """
        Method which applies the constraint

        :return: None if `no_set` is False, float otherwise.
        """
        if not self.enabled:
            if no_set:
                return None
            return
        independent_objs = None
        if isinstance(self.dependent_obj_ids, int):
            dependent_obj = self.get_obj(self.dependent_obj_ids)
        else:
            raise AttributeError
        if isinstance(self.independent_obj_ids, int):
            independent_objs = self.get_obj(self.independent_obj_ids)
        elif isinstance(self.independent_obj_ids, list):
            independent_objs = [
                self.get_obj(obj_id) for obj_id in self.independent_obj_ids
            ]
        if independent_objs is not None:
            value = self._parse_operator(independent_objs, *args, **kwargs)
        else:
            value = self._parse_operator(dependent_obj, *args, **kwargs)

        if not no_set:
            toggle = False
            if not dependent_obj.enabled:
                dependent_obj.enabled = True
                toggle = True
            dependent_obj.value = value
            if toggle:
                dependent_obj.enabled = False
        return value

    @abstractmethod
    def _parse_operator(self, obj: V, *args, **kwargs) -> Number:
        """
        Abstract method which contains the constraint logic

        :param obj: The object/objects which the constraint will use
        :return: A numeric result of the constraint logic
        """

    @abstractmethod
    def __repr__(self):
        pass

    def get_key(self, obj) -> int:
        """
        Get the unique key of a easyCore object

        :param obj: easyCore object
        :return: key for easyCore object
        """
        return self._borg.map.convert_id_to_key(obj)

    def get_obj(self, key: int) -> V:
        """
        Get an easyCore object from its unique key

        :param key: an easyCore objects unique key
        :return: easyCore object
        """
        return self._borg.map.get_item_by_key(key)


C = TypeVar("C", bound=ConstraintBase)


class NumericConstraint(ConstraintBase):
    """
    A `NumericConstraint` is a constraint whereby a dependent parameters value is something of an independent parameters
    value. I.e. a < 1, a > 5
    """

    def __init__(
        self, dependent_obj: V, operator: str, value: Number
    ):
        """
        A `NumericConstraint` is a constraint whereby a dependent parameters value is something of an independent
        parameters value. I.e. a < 1, a > 5

        :param dependent_obj: Dependent Parameter
        :param operator: Relation to between the parameter and the values. e.g. ``=``, ``<``, ``>``
        :param value: What the parameters value should be compared against.

        :example:
        .. code-block:: python

             from easyCore.Fitting.Constraints import NumericConstraint
             from easyCore.Objects.Base import Parameter
             # Create an `a < 1` constraint
             a = Parameter('a', 0.2)
             constraint = NumericConstraint(a, '<=', 1)
             a.user_constraints['LEQ_1'] = constraint
             # This works
             a.value = 0.85
             # This triggers the constraint
             a.value = 2.0
             # `a` is set to the maximum of the constraint (`a = 1`)
        """
        super(NumericConstraint, self).__init__(
            dependent_obj, operator=operator, value=value
        )

    def _parse_operator(
        self, obj: V, *args, **kwargs
    ) -> Number:
        value = obj.raw_value
        if isinstance(value, list):
            value = np.array(value)
        self.aeval.symtable["value1"] = value
        self.aeval.symtable["value2"] = self.value
        try:
            self.aeval.eval(f"value3 = value1 {self.operator} value2")
            logic = self.aeval.symtable["value3"]
            if isinstance(logic, np.ndarray):
                value[not logic] = self.aeval.symtable["value2"]
            else:
                if not logic:
                    value = self.aeval.symtable["value2"]
        except Exception as e:
            raise e
        finally:
            self.aeval.symtable.clear()
        return value

    def __repr__(self) -> str:
        return f"{self.__class__.__name__} with `value` {self.operator} {self.value}"


class SelfConstraint(ConstraintBase):
    """
    A `SelfConstraint` is a constraint which tests a logical constraint on a property of itself, similar to a
    `NumericConstraint`. i.e. a > a.min. These constraints are usually used in the internal easyCore logic.
    """

    def __init__(
        self, dependent_obj: V, operator: str, value: str
    ):
        """
        A `SelfConstraint` is a constraint which tests a logical constraint on a property of itself, similar to
        a `NumericConstraint`. i.e. a > a.min.

        :param dependent_obj: Dependent Parameter
        :param operator: Relation to between the parameter and the values. e.g. ``=``, ``<``, ``>``
        :param value: Name of attribute to be compared against

        :example:
        .. code-block:: python

             from easyCore.Fitting.Constraints import SelfConstraint
             from easyCore.Objects.Base import Parameter
             # Create an `a < a.max` constraint
             a = Parameter('a', 0.2, max=1)
             constraint = SelfConstraint(a, '<=', 'max')
             a.user_constraints['MAX'] = constraint
             # This works
             a.value = 0.85
             # This triggers the constraint
             a.value = 2.0
             # `a` is set to the maximum of the constraint (`a = 1`)
        """
        super(SelfConstraint, self).__init__(
            dependent_obj, operator=operator, value=value
        )

    def _parse_operator(
        self, obj: V, *args, **kwargs
    ) -> Number:
        value = obj.raw_value
        self.aeval.symtable["value1"] = value
        self.aeval.symtable["value2"] = getattr(obj, self.value)
        try:
            self.aeval.eval(f"value3 = value1 {self.operator} value2")
            logic = self.aeval.symtable["value3"]
            if isinstance(logic, np.ndarray):
                value[not logic] = self.aeval.symtable["value2"]
            else:
                if not logic:
                    value = self.aeval.symtable["value2"]
        except Exception as e:
            raise e
        finally:
            self.aeval.symtable.clear()
        return value

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__} with `value` {self.operator} obj.{self.value}"
        )


class ObjConstraint(ConstraintBase):
    """
    A `ObjConstraint` is a constraint whereby a dependent parameter is something of an independent parameter
    value. E.g. a (Dependent Parameter) = 2* b (Independent Parameter)
    """

    def __init__(
        self, dependent_obj: V, operator: str, independent_obj: V
    ):
        """
        A `ObjConstraint` is a constraint whereby a dependent parameter is something of an independent parameter
        value. E.g. a (Dependent Parameter) < b (Independent Parameter)

        :param dependent_obj: Dependent Parameter
        :param operator: Relation to between the independent parameter and dependent parameter. e.g. ``2 *``, ``1 +``
        :param independent_obj: Independent Parameter

        :example:
        .. code-block:: python

             from easyCore.Fitting.Constraints import ObjConstraint
             from easyCore.Objects.Base import Parameter
             # Create an `a = 2 * b` constraint
             a = Parameter('a', 0.2)
             b = Parameter('b', 1)

             constraint = ObjConstraint(a, '2*', b)
             b.user_constraints['SET_A'] = constraint
             b.value = 1
             # This triggers the constraint
             a.value # Should equal 2

        """
        super(ObjConstraint, self).__init__(
            dependent_obj, independent_obj=independent_obj, operator=operator
        )
        self.external = True

    def _parse_operator(
        self, obj: V, *args, **kwargs
    ) -> Number:
        value = obj.raw_value
        self.aeval.symtable["value1"] = value
        try:
            self.aeval.eval(f"value2 = {self.operator} value1")
            value = self.aeval.symtable["value2"]
        except Exception as e:
            raise e
        finally:
            self.aeval.symtable.clear()
        return value

    def __repr__(self) -> str:
        return f"{self.__class__.__name__} with `dependent_obj` = {self.operator} `independent_obj`"


class MultiObjConstraint(ConstraintBase):
    """
    A `MultiObjConstraint` is similar to :class:`easyCore.Fitting.Constraints.ObjConstraint` except that it relates to
    multiple independent objects.
    """

    def __init__(
        self,
        independent_objs: List[V],
        operator: List[str],
        dependent_obj: V,
        value: Number,
    ):
        """
        A `MultiObjConstraint` is similar to :class:`easyCore.Fitting.Constraints.ObjConstraint` except that it relates
        to one or more independent objects.

        E.g.
        * a (Dependent Parameter) + b (Independent Parameter) = 1
        * a (Dependent Parameter) + b (Independent Parameter) - 2*c (Independent Parameter) = 0

        :param independent_objs: List of Independent Parameters
        :param operator: List of operators operating on the Independent Parameters
        :param dependent_obj: Dependent Parameter
        :param value: Value of the expression

        :example:
        **a + b = 1**

        .. code-block:: python

             from easyCore.Fitting.Constraints import MultiObjConstraint
             from easyCore.Objects.Base import Parameter
             # Create an `a + b = 1` constraint
             a = Parameter('a', 0.2)
             b = Parameter('b', 0.3)

             constraint = MultiObjConstraint([b], ['+'], a, 1)
             b.user_constraints['SET_A'] = constraint
             b.value = 0.4
             # This triggers the constraint
             a.value # Should equal 0.6

        **a + b - 2c = 0**

        .. code-block:: python

             from easyCore.Fitting.Constraints import MultiObjConstraint
             from easyCore.Objects.Base import Parameter
             # Create an `a + b - 2c = 0` constraint
             a = Parameter('a', 0.5)
             b = Parameter('b', 0.3)
             c = Parameter('c', 0.1)

             constraint = MultiObjConstraint([b, c], ['+', '-2*'], a, 0)
             b.user_constraints['SET_A'] = constraint
             c.user_constraints['SET_A'] = constraint
             b.value = 0.4
             # This triggers the constraint. Or it could be triggered by changing the value of c
             a.value # Should equal 0.2

        .. note:: This constraint is evaluated as ``dependent`` = ``value`` - SUM(``operator_i`` ``independent_i``)
        """
        super(MultiObjConstraint, self).__init__(
            dependent_obj,
            independent_obj=independent_objs,
            operator=operator,
            value=value,
        )
        self.external = True

    def _parse_operator(
        self, independent_objs: List[V], *args, **kwargs
    ) -> Number:
        in_str = ""
        value = None
        for idx, obj in enumerate(independent_objs):
            self.aeval.symtable[
                "p" + str(self.independent_obj_ids[idx])
            ] = obj.raw_value
            in_str += " p" + str(self.independent_obj_ids[idx])
            if idx < len(self.operator):
                in_str += " " + self.operator[idx]
        try:
            self.aeval.eval(f"final_value = {self.value} - ({in_str})")
            value = self.aeval.symtable["final_value"]
        except Exception as e:
            raise e
        finally:
            self.aeval.symtable.clear()
        return value

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}"


class FunctionalConstraint(ConstraintBase):
    """
    Functional constraints do not depend on other parameters and as such can be more complex.
    """

    def __init__(
        self,
        dependent_obj: V,
        func: Callable,
        independent_objs: Optional[List[V]] = None,
    ):
        """
        Functional constraints do not depend on other parameters and as such can be more complex.

        :param dependent_obj: Dependent Parameter
        :param func: Function to be evaluated in the form ``f(value, *args, **kwargs)``

        :example:
        .. code-block:: python

            import numpy as np
            from easyCore.Fitting.Constraints import FunctionalConstraint
            from easyCore.Objects.Base import Parameter

            a = Parameter('a', 0.2, max=1)
            constraint = FunctionalConstraint(a, np.abs)

            a.user_constraints['abs'] = constraint

            # This triggers the constraint
            a.value = 0.85 # `a` is set to 0.85
            # This triggers the constraint
            a.value = -0.5 # `a` is set to 0.5
        """
        super(FunctionalConstraint, self).__init__(
            dependent_obj, independent_obj=independent_objs
        )
        self.function = func
        if independent_objs is not None:
            self.external = True

    def _parse_operator(
        self, obj: V, *args, **kwargs
    ) -> Number:
        self.aeval.symtable[f"f{id(self.function)}"] = self.function
        value_str = f"r_value = f{id(self.function)}("
        if isinstance(obj, list):
            for o in obj:
                value_str += f"{o.raw_value},"
            value_str = value_str[:-1]
        else:
            value_str += f"{obj.raw_value}"
        value_str += ")"
        try:
            self.aeval.eval(value_str)
            value = self.aeval.symtable["r_value"]
        except Exception as e:
            raise e
        finally:
            self.aeval.symtable.clear()
        return value

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}"


def cleanup_constraint(obj_id: str, enabled: bool):
    try:
        obj = borg.map.get_item_by_key(obj_id)
        obj.enabled = enabled
    except ValueError:
        if borg.debug:
            print(f"Object with ID {obj_id} has already been deleted")



#  SPDX-FileCopyrightText: 2022 easyCore contributors  <core@easyscience.software>
#  SPDX-License-Identifier: BSD-3-Clause
#  © 2021-2022 Contributors to the easyCore project <https://github.com/easyScience/easyCore>

__author__ = 'github.com/wardsimon'
__version__ = '0.1.0'

from typing import Callable, Union, TypeVar, List, Tuple, Any, Iterable

import weakref
# import pint_xarray
import xarray as xr

from easyCore import np, ureg
from easyCore.Fitting.fitting_template import FitResults

T_ = TypeVar('T_')


@xr.register_dataset_accessor("easyCore")
class easyCoreDatasetAccessor:
    """
    Accessor to extend an xarray DataSet to easyCore. These functions can be accessed by `obj.easyCore.func`.
    The objective for this class accessor is to facilitate:

- Creation of nd datasets by making and assigning axes and data more accessible
- To add and keep track of errors in the form of sigma for datasets.py
- To perform fitting on one or more data arrays in the dataset simultaneously.
- To facilitate dask computation if required.
    """

    def __init__(self, xarray_obj: xr.Dataset):
        """
        This is called whenever you access obj.easyCore, hence the attributes in the obj should only be written if they
        have not been previously instantiated.

        :param xarray_obj: DataSet which is called by obj.easyCore
        :type xarray_obj: xarray.Dataset
        """

        self._obj = xarray_obj
        self._core_object = None
        self.__error_mapper = {}
        self.sigma_label_prefix = 's_'
        if self._obj.attrs.get('name', None) is None:
            self._obj.attrs['name'] = ''
        if self._obj.attrs.get('description', None) is None:
            self._obj.attrs['description'] = ''
        if self._obj.attrs.get('url', None) is None:
            self._obj.attrs['url'] = ''
        if self._obj.attrs.get('units', None) is None:
            self._obj.attrs['units'] = {}

    @property
    def name(self) -> str:
        """
        Get the common name of the DataSet.

        :return: Common name of the DataSet
        :rtype: str
        """
        return self._obj.attrs['name']

    @name.setter
    def name(self, new_name: str):
        """
        Set the common name of the DataSet i.e could be experiment name...

        :param new_name: Common name of the DataSet
        :type new_name: str
        :return: None
        :rtype: None
        """
        self._obj.attrs['name'] = new_name

    @property
    def description(self) -> str:
        """
        Get a description of the DataSet

        :return: Description of the DataSet
        :rtype: str
        """
        return self._obj.attrs['description']

    @description.setter
    def description(self, new_description: str):
        """
        Set the description of the DataSet

        :param new_description: Description of the DataSet
        :type new_description: str
        :return: None
        :rtype: None
        """
        self._obj.attrs['description'] = new_description

    @property
    def url(self) -> str:
        """
        Get the url of the DataSet

        :return: URL of the DataSet (empty if no URL)
        :rtype: str
        """
        return self._obj.attrs['url']

    @url.setter
    def url(self, new_url: str):
        """
        Set the URL of the DataSet. This may be a DOI.

        :param new_url: New URL/DOI of the DataSet
        :type new_url: str
        :return:None
        :rtype: None
        """
        self._obj.attrs['url'] = new_url

    @property
    def core_object(self):
        """
        Get the core object associated to a DataSet. Note that this is called from a weakref. If the easyCore obj is
        garbage collected, None will be returned.

        :return: easyCore object associated with the DataSet
        :rtype: Any
        """
        if self._core_object is None:
            return None
        return self._core_object()

    @core_object.setter
    def core_object(self, new_core_object: Any):
        """
        Associate an easyCore object to a DataSet.

        :param new_core_object: easyCore object to be associated to the DataSet
        :type new_core_object: Any
        :return: None
        :rtype: None
        """
        self._core_object = weakref.ref(new_core_object)

    def add_coordinate(self, coordinate_name: str, coordinate_values: Union[List[T_], np.ndarray], unit: str = ''):
        """
        Add a coordinate to the DataSet. This can be then be assigned to one or more DataArrays.

        :param coordinate_name: Name of the coordinate e.g. `x`
        :type coordinate_name: str
        :param coordinate_values: Points for the coordinates
        :type coordinate_values: Union[List[T_], numpy.ndarray]
        :param unit: Unit associated with the coordinate
        :type unit: str
        :return: None
        :rtype: None
        """
        self._obj.coords[coordinate_name] = coordinate_values
        self._obj.attrs['units'][coordinate_name] = ureg.Unit(unit)

    def remove_coordinate(self, coordinate_name: str):
        """
        Remove a coordinate from the DataSet. Note that this will not remove the coordinate from DataArrays which have
        already used the it!
        
        :param coordinate_name: Name of the coordinate to be removed 
        :type coordinate_name: str
        :return: None
        :rtype: None
        """
        del self._obj.coords[coordinate_name]
        del self._obj.attrs['units'][coordinate_name]

    def add_variable(self, variable_name, variable_coordinates: Union[str, List[str]],
                     variable_values: Union[List[T_], np.ndarray], variable_sigma: Union[List[T_], np.ndarray] = None,
                     unit: str = '', auto_sigma: bool = False):
        """
        Create a DataArray from known coordinates and data, assign it to the dataset under a given name. Variances can
        be calculated assuming gaussian distribution to 1 sigma.
        
        :param variable_name: Name of the DataArray which will be created and added to the dataset 
        :type variable_name: str
        :param variable_coordinates: List of coordinates used in the supplied data array. 
        :type variable_coordinates: str, List[str]
        :param variable_values: Numpy or list of data which will be assigned to the DataArray
        :type variable_values: Union[numpy.ndarray, list]
        :param variable_sigma: If the sigmas of the dataset are known, they can be supplied here.
        :type variable_sigma: Union[numpy.ndarray, list]
        :param unit: Unit associated with the DataArray
        :type unit: str
        :param auto_sigma: Should the sigma DataArray be automatically calculated assuming gaussian probability? 
        :type auto_sigma: bool
        :return: None
        :rtype: None
        """

        # Check if a user has supplied a coordinate as a string. Make it a list of strings
        if isinstance(variable_coordinates, str):
            variable_coordinates = [variable_coordinates]

        # The variable_coordinates can be any iterable object. Though we would assume list/tuple
        if not isinstance(variable_coordinates, Iterable):
            raise ValueError('The variable coordinates must be a list of strings')

        # Check to see if the user want to assign a coordinate which does not exist yet.
        known_keys = self._obj.coords.keys()
        for dimension in variable_coordinates:
            if dimension not in known_keys:
                raise ValueError(f'The supplied coordinate `{dimension}` must first be defined.')

        # Create  the dataset.
        self._obj[variable_name] = (variable_coordinates, variable_values)

        # Deal with sigmas
        if variable_sigma is not None:
            # CASE 1, user has supplied sigmas
            if isinstance(variable_sigma, Callable):
                # CASE 1-1, The sigmas are created by some kind of generator
                self.sigma_generator(variable_name, variable_sigma)
            elif isinstance(variable_sigma, np.ndarray):
                # CASE 1-2, The sigmas are a numpy arrays
                self.sigma_attach(variable_name, variable_sigma)
            elif isinstance(variable_sigma, list):
                # CASE 1-3, We have been given a list. Make it a numpy array
                self.sigma_attach(variable_name, np.array(variable_sigma))
            else:
                raise ValueError('User supplied sigmas must be of the form; Callable fn, numpy array, list')
        else:
            # CASE 2, No sigmas have been supplied.
            if auto_sigma:
                # CASE 2-1, Automatically generate the sigmas using gaussian probability
                self.sigma_generator(variable_name)

        # Set units for the newly created DataArray
        self._obj.attrs['units'][variable_name] = ureg.Unit(unit)
        # If a sigma has been attached, attempt to work out the units.
        if unit and variable_sigma is None and auto_sigma:
            self._obj.attrs['units'][self.sigma_label_prefix + variable_name] = ureg.Unit(unit + ' ** 0.5')
        else:
            if auto_sigma:
                self._obj.attrs['units'][self.sigma_label_prefix + variable_name] = ureg.Unit('')

    def remove_variable(self, variable_name: str):
        """
        Remove a DataArray from the DataSet by supplied name.

        :param variable_name: Name of DataArray to be removed
        :type variable_name: str
        :return: None
        :rtype: None
        """
        del self._obj[variable_name]

    def sigma_generator(self, variable_label: str,
                        sigma_func: Callable = lambda x: np.sqrt(np.abs(x)),
                        label_prefix: str = None):
        """
        Generate sigmas off of a DataArray based on a function.

        :param variable_label: Name of the DataArray to perform the calculation on
        :type variable_label: str
        :param sigma_func: Function to generate the sigmas. Must be of the form f(x) and return an array of the same
        shape as the input. DEFAULT: sqrt(|data|)
        :type sigma_func: Callable
        :param label_prefix: What prefix should be used to designate a sigma DataArray from a data DataArray
        :type label_prefix: str
        :return: None
        :rtype: None
        """
        sigma_values = sigma_func(self._obj[variable_label])
        self.sigma_attach(variable_label, sigma_values, label_prefix)

    def sigma_attach(self, variable_label: str,
                     sigma_values: Union[List[T_], np.ndarray, xr.DataArray],
                     label_prefix: str = None):
        """
        Attach an array of sigmas to the DataSet.

        :param variable_label: Name of the DataArray to perform the calculation on
        :type variable_label: str
        :param sigma_values: Array of sigmas in list, numpy or DataArray form
        :type sigma_values: Union[List[T_], numpy.ndarray, xarray.DataArray]
        :param label_prefix: What prefix should be used to designate a sigma DataArray from a data DataArray
        :type label_prefix: str
        :return: None
        :rtype: None
        """
        # Use the default sigma prefix if not defined.
        if label_prefix is None:
            label_prefix = self.sigma_label_prefix

        # Form the label for the new DataArray
        sigma_label = label_prefix + variable_label

        # Map the original DataArray to the new sigma DataArray
        self.__error_mapper[variable_label] = sigma_label
        # Assign the sigma DataArray to the DataSet
        if not isinstance(sigma_values, xr.DataArray):
            self._obj[sigma_label] = (list(self._obj[variable_label].coords.keys()), sigma_values)
        else:
            self._obj[sigma_label] = sigma_values

    def generate_points(self, coordinates: List[str]) -> xr.DataArray:
        """
        Generate an expanded DataArray of points which corresponds to broadcasted dimensions (`all_x`) which have been
        concatenated along the second axis (`fit_dim`).

        :param coordinates: List of coordinate names to broadcast and concatenate along
        :type coordinates: List[str]
        :return: Broadcasted and concatenated coordinates
        :rtype: xarray.DataArray

.. code-block:: python

     x = [1, 2], y = [3, 4]
     d = xr.DataArray()
     d.easyCore.add_coordinate('x', x)
     d.easyCore.add_coordinate('y', y)
     points = d.easyCore.generate_points(['x', 'y'])
     print(points)
        """

        coords = [self._obj.coords[da] for da in coordinates]
        c_array = []
        n_array = []
        for da in xr.broadcast(*coords):
            c_array.append(da)
            n_array.append(da.name)

        f = xr.concat(c_array, dim='fit_dim')
        f = f.stack(all_x=n_array)
        return f

    def fit(self, fitter, data_arrays: list, *args,
            dask: str = 'forbidden',
            fit_kwargs: dict = None,
            fn_kwargs: dict = None,
            vectorized: bool = False,
            **kwargs) -> List[FitResults]:
        """
        Perform a fit on one or more DataArrays. This fit utilises a given fitter from `easyCore.Fitting.Fitter`, though
        there are a few differences to a standard easyCore fit. In particular, key-word arguments to control the
        optimisation algorithm go in the `fit_kwargs` dictionary, fit function key-word arguments go in the `fn_kwargs`
        and given key-word arguments control the `xarray.apply_ufunc` function.

        :param fitter: Fitting object which controls the fitting
        :type fitter: easyCore.Fitting.Fitter
        :param args: Arguments to go to the fit function
        :type args: Any
        :param dask: Dask control string. See `xarray.apply_ufunc` documentation
        :type dask: str
        :param fit_kwargs: Dictionary of key-word arguments to be supplied to the Fitting control
        :type fit_kwargs: dict
        :param fn_kwargs: Dictionary of key-words to be supplied to the fit function
        :type fn_kwargs: dict
        :param vectorized: Should the fit function be given dependents in a single object or split
        :type vectorized: bool
        :param kwargs: Key-word arguments for `xarray.apply_ufunc`. See `xarray.apply_ufunc` documentation
        :type kwargs: Any
        :return: Results of the fit
        :rtype: List[FitResults]
        """

        if fn_kwargs is None:
            fn_kwargs = {}
        if fit_kwargs is None:
            fit_kwargs = {}
        if not isinstance(data_arrays, (list, tuple)):
            data_arrays = [data_arrays]

        # In this case we are only fitting 1 dataset
        if len(data_arrays) == 1:
            variable_label = data_arrays[0]
            dataset = self._obj[variable_label]
            if self.__error_mapper.get(variable_label, False):
                # Pull out any sigmas and send them to the fitter.
                temp = self._obj[self.__error_mapper[variable_label]]
                temp[xr.ufuncs.isnan(temp)] = 1e5
                fit_kwargs['weights'] = temp
            # Perform a standard DataArray fit.
            return dataset.easyCore.fit(fitter, *args,
                                        fit_kwargs=fit_kwargs,
                                        fn_kwargs=fn_kwargs,
                                        dask=dask,
                                        vectorize=vectorized,
                                        **kwargs)
        else:
            # In this case we are fitting multiple datasets to the same fn!
            bdim_f = [self._obj[p].easyCore.fit_prep(fitter.fit_function) for p in data_arrays]
            dim_names = [list(self._obj[p].dims.keys()) if isinstance(self._obj[p].dims, dict) else self._obj[p].dims for p in data_arrays]
            bdims = [bdim[0] for bdim in bdim_f]
            fs = [bdim[1] for bdim in bdim_f]
            old_fit_func = fitter.fit_function

            fn_array = []
            y_list = []
            for _idx, d in enumerate(bdims):
                dims = self._obj[data_arrays[_idx]].dims
                if isinstance(dims, dict):
                    dims = list(dims.keys())

                def local_fit_func(x, *args, idx=None, **kwargs):
                    kwargs['vectorize'] = vectorized
                    res = xr.apply_ufunc(fs[idx], *bdims[idx], *args, dask=dask, kwargs=fn_kwargs, **kwargs)
                    if dask != 'forbidden':
                        res.compute()
                    return res.stack(all_x=dim_names[idx])
                y_list.append(self._obj[data_arrays[_idx]].stack(all_x=dims))
                fn_array.append(local_fit_func)

            def fit_func(x, *args, **kwargs):
                res = []
                for idx in range(len(fn_array)):
                    res.append(fn_array[idx](x, *args, idx=idx, **kwargs))
                return xr.DataArray(np.concatenate(res, axis=0), coords={'all_x': x}, dims='all_x')

            fitter.initialize(fitter.fit_object, fit_func)
            try:
                if fit_kwargs.get('weights', None) is not None:
                    del fit_kwargs['weights']
                x = xr.DataArray(np.arange(np.sum([y.size for y in y_list])), dims='all_x')
                y = xr.DataArray(np.concatenate(y_list, axis=0), coords={'all_x': x}, dims='all_x')
                f_res = fitter.fit(x, y, **fit_kwargs)
                f_res = check_sanity_multiple(f_res, [self._obj[p] for p in data_arrays])
            finally:
                fitter.fit_function = old_fit_func
            return f_res


@xr.register_dataarray_accessor("easyCore")
class easyCoreDataarrayAccessor:
    """
    Accessor to extend an xarray DataArray to easyCore. These functions can be accessed by `obj.easyCore.func`.

    """

    def __init__(self, xarray_obj: xr.DataArray):
        self._obj = xarray_obj
        self._core_object = None
        self.sigma_label_prefix = 's_'
        if self._obj.attrs.get('computation', None) is None:
            self._obj.attrs['computation'] = {
                'precompute_func':  None,
                'compute_func': None,
                'postcompute_func': None
            }

    def __empty_functional(self) -> Callable:

        def outer():
            def empty_fn(input, *args, **kwargs):
                return input
            return empty_fn

        class wrapper:
            def __init__(obj):
                obj.obj = self
                obj.data = {}
                obj.fn = outer()

            def __call__(self, *args, **kwargs):
                return self.fn(*args, **kwargs)

        return wrapper()

    @property
    def core_object(self):
        """
        Get the core object associated to a DataArray. Note that this is called from a weakref. If the easyCore obj is
        garbage collected, None will be returned.

        :return: easyCore object associated with the DataArray
        :rtype: Any
        """
        if self._core_object is None:
            return None
        return self._core_object()

    @core_object.setter
    def core_object(self, new_core_object: Any):
        """
        Set the core object associated to a dataset

        :param new_core_object: easyCore object to be associated with the DataArray
        :type new_core_object: Any
        :return: None
        :rtype: None
        """
        self._core_object = weakref.ref(new_core_object)

    @property
    def compute_func(self) -> Callable:
        """
        Get the computational function which will be executed during a fit

        :return: Computational function applied to the DataArray
        :rtype: Callable
        """
        result = self._obj.attrs['computation']['compute_func']
        if result is None:
            result = self.__empty_functional()
        return result

    @compute_func.setter
    def compute_func(self, new_computational_fn: Callable):
        """
        Set the computational function which is called during a fit

        :param new_computational_fn: Computational function applied to the DataArray
        :type new_computational_fn: Callable
        :return: None
        :rtype: None
        """
        self._obj.attrs['computation']['compute_func'] = new_computational_fn

    @property
    def precompute_func(self) -> Callable:
        """
        Get the pre-computational function which will be executed before a fit

        :return: Computational function applied to the DataArray before fitting
        :rtype: Callable
        """
        result = self._obj.attrs['computation']['precompute_func']
        if result is None:
            result = self.__empty_functional()
        return result

    @precompute_func.setter
    def precompute_func(self, new_computational_fn: Callable):
        """
        Set the computational function which is called before a fit

        :param new_computational_fn: Computational function applied to the DataArray before fitting
        :type new_computational_fn: Callable
        :return: None
        :rtype: None
        """
        self._obj.attrs['computation']['precompute_func'] = new_computational_fn

    @property
    def postcompute_func(self) -> Callable:
        """
        Get the post-computational function which will be executed after a fit

        :return: Computational function applied to the DataArray after fitting
        :rtype: Callable
        """
        result = self._obj.attrs['computation']['postcompute_func']
        if result is None:
            result = self.__empty_functional()
        return result

    @postcompute_func.setter
    def postcompute_func(self, new_computational_fn: Callable):
        """
        Set the computational function which is called after a fit

        :param new_computational_fn: Computational function applied to the DataArray after fitting
        :type new_computational_fn: Callable
        :return: None
        :rtype: None
        """
        self._obj.attrs['computation']['postcompute_func'] = new_computational_fn

    def fit_prep(self, func_in: Callable, bdims=None, dask_chunks=None) -> Tuple[xr.DataArray, Callable]:
        """
        Generate boradcasted coordinates for fitting and reform the fitting function into one which can handle xarrays

        :param func_in: Function to be wrapped and made xarray fitting compatable.
        :type func_in: Callable
        :param bdims: Optional precomputed broadcasted dimensions.
        :type bdims: xarray.DataArray
        :param dask_chunks: How to split to broadcasted dimensions for dask.
        :type dask_chunks: Tuple[int..]
        :return: Tuple of broadcasted fit arrays and wrapped fit function.
        :rtype: xarray.DataArray, Callable
        """

        if bdims is None:
            coords = [self._obj.coords[da].transpose() for da in self._obj.dims]
            bdims = xr.broadcast(*coords)
        self._obj.attrs['computation']['compute_func'] = func_in

        def func(x, *args, vectorize: bool = False, **kwargs):
            old_shape = x.shape
            if not vectorize:
                xs = [x_new.flatten() for x_new in [x, *args] if isinstance(x_new, np.ndarray)]
                x_new = np.column_stack(xs)
                if len(x_new.shape) > 1 and x_new.shape[1] == 1:
                    x_new = x_new.reshape((-1))
                result = self.compute_func(x_new, **kwargs)
            else:
                result = self.compute_func(*[d for d in [x, args] if isinstance(d, np.ndarray)],
                                 *[d for d in args if not isinstance(d, np.ndarray)],
                                 **kwargs)
            if isinstance(result, np.ndarray):
                result = result.reshape(old_shape)
            result = self.postcompute_func(result)
            return result

        return bdims, func

    def generate_points(self) -> xr.DataArray:
        """
        Generate an expanded DataArray of points which corresponds to broadcasted dimensions (`all_x`) which have been
        concatenated along the second axis (`fit_dim`).

        :return: Broadcasted and concatenated coordinates
        :rtype: xarray.DataArray
        """

        coords = [self._obj.coords[da] for da in self._obj.dims]
        c_array = []
        n_array = []
        for da in xr.broadcast(*coords):
            c_array.append(da)
            n_array.append(da.name)

        f = xr.concat(c_array, dim='fit_dim')
        f = f.stack(all_x=n_array)
        return f

    def fit(self, fitter, *args,
            fit_kwargs: dict = None,
            fn_kwargs: dict = None,
            vectorize: bool = False,
            dask: str = 'forbidden',
            **kwargs) -> FitResults:
        """
        Perform a fit on the given DataArray. This fit utilises a given fitter from `easyCore.Fitting.Fitter`, though
        there are a few differences to a standard easyCore fit. In particular, key-word arguments to control the
        optimisation algorithm go in the `fit_kwargs` dictionary, fit function key-word arguments go in the `fn_kwargs`
        and given key-word arguments control the `xarray.apply_ufunc` function.

        :param fitter: Fitting object which controls the fitting
        :type fitter: easyCore.Fitting.Fitter
        :param args: Arguments to go to the fit function
        :type args: Any
        :param dask: Dask control string. See `xarray.apply_ufunc` documentation
        :type dask: str
        :param fit_kwargs: Dictionary of key-word arguments to be supplied to the Fitting control
        :type fit_kwargs: dict
        :param fn_kwargs: Dictionary of key-words to be supplied to the fit function
        :type fn_kwargs: dict
        :param vectorize: Should the fit function be given dependents in a single object or split
        :type vectorize: bool
        :param kwargs: Key-word arguments for `xarray.apply_ufunc`. See `xarray.apply_ufunc` documentation
        :type kwargs: Any
        :return: Results of the fit
        :rtype: FitResults
        """

        # Deal with any kwargs which has been given
        if fn_kwargs is None:
            fn_kwargs = {}
        if fit_kwargs is None:
            fit_kwargs = {}
        old_fit_func = fitter.fit_function

        # Wrap and broadcast
        bdims, f = self.fit_prep(fitter.fit_function)
        dims = self._obj.dims

        # Find which coords we need
        if isinstance(dims, dict):
            dims = list(dims.keys())

        # Wrap the wrap in a callable
        def local_fit_func(x, *args, **kwargs):
            """
            Function which will be called by the fitter. This will deal with sending the function the correct data.
            """
            kwargs['vectorize'] = vectorize
            res = xr.apply_ufunc(f, *bdims, *args, dask=dask, kwargs=fn_kwargs, **kwargs)
            if dask != 'forbidden':
                res.compute()
            return res.stack(all_x=dims)

        # Set the new callable to the fitter and initialize
        fitter.initialize(fitter.fit_object, local_fit_func)
        # Make easyCore.Fitting.Fitter compatible `x`
        x_for_fit = xr.concat(bdims, dim='fit_dim')
        x_for_fit = x_for_fit.stack(all_x=[d.name for d in bdims])
        try:
            # Deal with any sigmas if supplied
            if fit_kwargs.get('weights', None) is not None:
                fit_kwargs['weights'] = xr.DataArray(
                    np.array(fit_kwargs['weights']),
                    dims=['all_x'],
                    coords={'all_x': x_for_fit.all_x}
                )
            # Try to perform a fit
            f_res = fitter.fit(x_for_fit, self._obj.stack(all_x=dims), **fit_kwargs)
            f_res = check_sanity_single(f_res)
        finally:
            # Reset the fit function on the fitter to the old fit function.
            fitter.fit_function = old_fit_func
        return f_res


def check_sanity_single(fit_results: FitResults) -> FitResults:
    """
    Convert the FitResults from a fitter compatible state to a recognizable DataArray state.

    :param fit_results: Results of a fit to be modified
    :type fit_results: FitResults
    :return: Modified fit results
    :rtype: FitResults
    """
    items = ['y_obs', 'y_calc', 'residual']

    for item in items:
        array = getattr(fit_results, item)
        if isinstance(array, xr.DataArray):
            array = array.unstack()
            array.name = item
            setattr(fit_results, item, array)

    x_array = fit_results.x
    if isinstance(x_array, xr.DataArray):
        fit_results.x.name = 'axes_broadcast'
        x_array = x_array.unstack()
        x_dataset = xr.Dataset()
        dims = [dims for dims in x_array.dims if dims != 'fit_dim']
        for idx, dim in enumerate(dims):
            x_dataset[dim + '_broadcast'] = x_array[idx]
            x_dataset[dim + '_broadcast'].name = dim + '_broadcast'
        fit_results.x_matrices = x_dataset
    else:
        fit_results.x_matrices = x_array
    return fit_results


def check_sanity_multiple(fit_results: FitResults, originals: List[xr.DataArray]) -> List[FitResults]:
    """
    Convert the multifit FitResults from a fitter compatible state to a list of recognizable DataArray states.

    :param fit_results: Results of a fit to be modified
    :type fit_results: FitResults
    :param originals: List of DataArrays which were fitted against, so we can resize and re-chunk the results
    :type originals: List[xr.DataArray]
    :return: Modified fit results
    :rtype: List[FitResults]
    """

    return_results = []
    offset = 0
    for item in originals:
        current_results = fit_results.__class__()
        # Fill out the basic stuff....
        current_results.engine_result = fit_results.engine_result
        current_results.fitting_engine = fit_results.fitting_engine
        current_results.success = fit_results.success
        current_results.p = fit_results.p
        current_results.p0 = fit_results.p0
        # now the tricky stuff
        current_results.x = item.easyCore.generate_points()
        current_results.y_obs = item.copy()
        current_results.y_obs.name = f'{item.name}_obs'
        current_results.y_calc = xr.DataArray(
            fit_results.y_calc[offset:offset+item.size].data,
            dims=item.dims,
            coords=item.coords,
            name=f'{item.name}_calc'
        )
        offset += item.size
        current_results.residual = current_results.y_calc - current_results.y_obs
        current_results.residual.name = f'{item.name}_residual'
        return_results.append(current_results)
    return return_results



#  SPDX-FileCopyrightText: 2022 easyCore contributors  <core@easyscience.software>
#  SPDX-License-Identifier: BSD-3-Clause
#  © 2021-2022 Contributors to the easyCore project <https://github.com/easyScience/easyCore>

__author__ = "github.com/wardsimon"
__version__ = "0.1.0"

import inspect
from typing import List, Optional

from easyCore.Fitting.fitting_template import (
    Union,
    Callable,
    FittingTemplate,
    np,
    FitResults,
    NameConverter,
    FitError,
)

# Import bumps specific objects
from bumps.names import Curve, FitProblem
from bumps.parameter import Parameter as bumpsParameter
from bumps.fitters import fit as bumps_fit, FIT_AVAILABLE_IDS


class bumps(FittingTemplate):  # noqa: S101
    """
    This is a wrapper to bumps: https://bumps.readthedocs.io/
    It allows for the bumps fitting engine to use parameters declared in an `easyCore.Objects.Base.BaseObj`.
    """

    property_type = bumpsParameter
    name = "bumps"

    def __init__(self, obj, fit_function: Callable):
        """
        Initialize the fitting engine with a `BaseObj` and an arbitrary fitting function.

        :param obj: Object containing elements of the `Parameter` class
        :type obj: BaseObj
        :param fit_function: function that when called returns y values. 'x' must be the first
                            and only positional argument. Additional values can be supplied by
                            keyword/value pairs
        :type fit_function: Callable
        """
        super().__init__(obj, fit_function)
        self._cached_pars_order = ()
        self.p_0 = {}

    def make_model(
        self, pars: Optional[List[bumpsParameter]] = None
    ) -> Callable:
        """
        Generate a bumps model from the supplied `fit_function` and parameters in the base object.
        Note that this makes a callable as it needs to be initialized with *x*, *y*, *weights*

        :return: Callable to make a bumps Curve model
        :rtype: Callable
        """
        fit_func = self._generate_fit_function()

        def outer(obj):
            def make_func(x, y, weights):
                par = {}
                if not pars:
                    for name, item in obj._cached_pars.items():
                        par["p" + str(name)] = obj.convert_to_par_object(item)
                else:
                    for item in pars:
                        par[
                            "p" + str(NameConverter().get_key(item))
                        ] = obj.convert_to_par_object(item)
                return Curve(fit_func, x, y, dy=weights, **par)

            return make_func

        return outer(self)

    def _generate_fit_function(self) -> Callable:
        """
        Using the user supplied `fit_function`, wrap it in such a way we can update `Parameter` on
        iterations.

        :return: a fit function which is compatible with bumps models
        :rtype: Callable
        """
        # Original fit function
        func = self._original_fit_function
        # Get a list of `Parameters`
        self._cached_pars = {}
        for parameter in self._object.get_fit_parameters():
            self._cached_pars[NameConverter().get_key(parameter)] = parameter

        # Make a new fit function
        def fit_function(x: np.ndarray, **kwargs):
            """
            Wrapped fit function which now has a bumps compatible form

            :param x: array of data points to be calculated
            :type x: np.ndarray
            :param kwargs: key word arguments
            :return: points calculated at `x`
            :rtype: np.ndarray
            """
            # Update the `Parameter` values and the callback if needed
            for name, value in kwargs.items():
                par_name = int(name[1:])
                if par_name in self._cached_pars.keys():
                    self._cached_pars[par_name].value = value
                    update_fun = self._cached_pars[par_name]._callback.fset
                    if update_fun:
                        update_fun(value)
            # TODO Pre processing here
            for constraint in self.fit_constraints():
                constraint()
            return_data = func(x)
            # TODO Loading or manipulating data here
            return return_data

        # Fake the function signature.
        # This is done as lmfit wants the function to be in the form:
        # f = (x, a=1, b=2)...
        # Where we need to be generic. Note that this won't hold for much outside of this scope.

        self._cached_pars_order = tuple(self._cached_pars.keys())
        params = [
            inspect.Parameter(
                "x", inspect.Parameter.POSITIONAL_OR_KEYWORD, annotation=inspect._empty
            ),
            *[
                inspect.Parameter(
                    "p" + str(name),
                    inspect.Parameter.POSITIONAL_OR_KEYWORD,
                    annotation=inspect._empty,
                    default=self._cached_pars[name].raw_value,
                )
                for name in self._cached_pars_order
            ],
        ]
        # Sign the function
        fit_function.__signature__ = inspect.Signature(params)
        self._fit_function = fit_function
        return fit_function

    def fit(
        self,
        x: np.ndarray,
        y: np.ndarray,
        weights: Optional[np.ndarray] = None,
        model: Optional = None,
        parameters: Optional = None,
        method: Optional[str] = None,
        minimizer_kwargs: Optional[dict] = None,
        engine_kwargs: Optional[dict] = None,
        **kwargs,
    ) -> FitResults:
        """
        Perform a fit using the lmfit engine.

        :param x: points to be calculated at
        :type x: np.ndarray
        :param y: measured points
        :type y: np.ndarray
        :param weights: Weights for supplied measured points * Not really optional*
        :type weights: np.ndarray
        :param model: Optional Model which is being fitted to
        :type model: lmModel
        :param parameters: Optional parameters for the fit
        :type parameters: List[bumpsParameter]
        :param kwargs: Additional arguments for the fitting function.
        :param method: Method for minimization
        :type method: str
        :return: Fit results
        :rtype: ModelResult
        """

        default_method = {}
        if method is not None and method in self.available_methods():
            default_method["method"] = method

        if weights is None:
            weights = np.sqrt(np.abs(y))

        if engine_kwargs is None:
            engine_kwargs = {}

        if minimizer_kwargs is None:
            minimizer_kwargs = {}
        # else:
        #     minimizer_kwargs = {"fit_kws": minimizer_kwargs}
        minimizer_kwargs.update(engine_kwargs)

        if model is None:
            model = self.make_model(pars=parameters)
            model = model(x, y, weights)
        self._cached_model = model
        self.p_0 = {
            f"p{key}": self._cached_pars[key].raw_value
            for key in self._cached_pars.keys()
        }
        problem = FitProblem(model)
        # Why do we do this? Because a fitting template has to have borg instantiated outside pre-runtime
        from easyCore import borg

        borg.stack.beginMacro("Fitting routine")
        try:
            model_results = bumps_fit(
                problem, **default_method, **minimizer_kwargs, **kwargs
            )
            self._set_parameter_fit_result(model_results)
            results = self._gen_fit_results(model_results)
        except Exception as e:
            raise FitError(e)
        finally:
            borg.stack.endMacro()
        return results

    def convert_to_pars_obj(
        self, par_list: Optional[List] = None
    ) -> List[bumpsParameter]:
        """
        Create a container with the `Parameters` converted from the base object.

        :param par_list: If only a single/selection of parameter is required. Specify as a list
        :type par_list: List[str]
        :return: bumps Parameters list
        :rtype: List[bumpsParameter]
        """
        if par_list is None:
            # Assume that we have a BaseObj for which we can obtain a list
            par_list = self._object.get_fit_parameters()
        pars_obj = [self.__class__.convert_to_par_object(obj) for obj in par_list]
        return pars_obj

    # For some reason I have to double staticmethod :-/
    @staticmethod
    def convert_to_par_object(obj) -> bumpsParameter:
        """
        Convert an `easyCore.Objects.Base.Parameter` object to a bumps Parameter object

        :return: bumps Parameter compatible object.
        :rtype: bumpsParameter
        """
        return bumpsParameter(
            name="p" + str(NameConverter().get_key(obj)),
            value=obj.raw_value,
            bounds=[obj.min, obj.max],
            fixed=obj.fixed,
        )

    def _set_parameter_fit_result(self, fit_result):
        """
        Update parameters to their final values and assign a std error to them.

        :param fit_result: Fit object which contains info on the fit
        :return: None
        :rtype: noneType
        """
        pars = self._cached_pars
        for index, name in enumerate(self._cached_model._pnames):
            dict_name = int(name[1:])
            pars[dict_name].value = fit_result.x[index]
            pars[dict_name].error = fit_result.dx[index]

    def _gen_fit_results(self, fit_results, **kwargs) -> FitResults:
        """
        Convert fit results into the unified `FitResults` format

        :param fit_result: Fit object which contains info on the fit
        :return: fit results container
        :rtype: FitResults
        """

        results = FitResults()
        for name, value in kwargs.items():
            if getattr(results, name, False):
                setattr(results, name, value)
        results.success = fit_results.success
        pars = self._cached_pars
        item = {}
        for index, name in enumerate(self._cached_model._pnames):
            dict_name = int(name[1:])
            item[name] = pars[dict_name].raw_value
        results.p0 = self.p_0
        results.p = item
        results.x = self._cached_model.x
        results.y_obs = self._cached_model.y
        results.y_calc = self.evaluate(results.x, parameters=results.p)
        results.residual = results.y_obs - results.y_calc
        results.goodness_of_fit = fit_results.fun

        results.fitting_engine = self.__class__
        results.fit_args = None
        results.engine_result = fit_results
        # results.check_sanity()
        return results

    def available_methods(self) -> List[str]:
        return FIT_AVAILABLE_IDS



#  SPDX-FileCopyrightText: 2022 easyCore contributors  <core@easyscience.software>
#  SPDX-License-Identifier: BSD-3-Clause
#  © 2021-2022 Contributors to the easyCore project <https://github.com/easyScience/easyCore>

from __future__ import annotations

__author__ = "github.com/wardsimon"
__version__ = "0.1.0"

from numbers import Number
from typing import Union, TypeVar, Optional, TYPE_CHECKING, Callable, List, Dict, Any, Tuple

from easyCore import borg
from easyCore.Objects.ObjectClasses import BasedBase, Descriptor
from collections.abc import MutableSequence
from easyCore.Utils.UndoRedo import NotarizedDict

if TYPE_CHECKING:
    from easyCore.Utils.typing import B, iF, V


class BaseCollection(BasedBase, MutableSequence):
    """
    This is the base class for which all higher level classes are built off of.
    NOTE: This object is serializable only if parameters are supplied as:
    `BaseObj(a=value, b=value)`. For `Parameter` or `Descriptor` objects we can
    cheat with `BaseObj(*[Descriptor(...), Parameter(...), ...])`.
    """

    def __init__(
        self,
        name: str,
        *args: Union[B, V],
        interface: Optional[iF] = None,
        **kwargs,
    ):
        """
        Set up the base collection class.

        :param name: Name of this object
        :type name: str
        :param args: selection of
        :param _kwargs: Fields which this class should contain
        :type _kwargs: dict
        """
        BasedBase.__init__(self, name)
        kwargs = {key: kwargs[key] for key in kwargs.keys() if kwargs[key] is not None}

        for item in [*kwargs.values(), *args]:
            if not issubclass(type(item), (Descriptor, BasedBase)):
                raise AttributeError(
                    "A collection can only be formed from easyCore objects."
                )

        _kwargs = {}
        for key, item in kwargs.items():
            _kwargs[key] = item
        for arg in args:
            kwargs[str(borg.map.convert_id_to_key(arg))] = arg
            _kwargs[str(borg.map.convert_id_to_key(arg))] = arg

        # Set kwargs, also useful for serialization
        self._kwargs = NotarizedDict(**_kwargs)

        for key in kwargs.keys():
            if key in self.__dict__.keys() or key in self.__slots__:
                raise AttributeError(
                    f"Given kwarg: `{key}`, is an internal attribute. Please rename."
                )
            self._borg.map.add_edge(self, kwargs[key])
            self._borg.map.reset_type(kwargs[key], "created_internal")
            if interface is not None:
                kwargs[key].interface = interface
            # TODO wrap getter and setter in Logger
        if interface is not None:
            self.interface = interface
        self._kwargs._stack_enabled = True

    def insert(self, index: int, value: Union[V, B]) -> None:
        """
        Insert an object into the collection at an index.

        :param index: Index for easyCore object to be inserted.
        :type index: int
        :param value: Object to be inserted.
        :type value: Union[BasedBase, Descriptor]
        :return: None
        :rtype: None
        """
        t_ = type(value)
        if issubclass(t_, (BasedBase, Descriptor)):
            update_key = list(self._kwargs.keys())
            values = list(self._kwargs.values())
            # Update the internal dict
            new_key = str(borg.map.convert_id_to_key(value))
            update_key.insert(index, new_key)
            values.insert(index, value)
            self._kwargs.reorder(**{k: v for k, v in zip(update_key, values)})
            # ADD EDGE
            self._borg.map.add_edge(self, value)
            self._borg.map.reset_type(value, "created_internal")
            value.interface = self.interface
        else:
            raise AttributeError(
                "Only easyCore objects can be put into an easyCore group"
            )

    def __getitem__(self, idx: Union[int, slice]) -> Union[V, B]:
        """
        Get an item in the collection based on it's index.

        :param idx: index or slice of the collection.
        :type idx: Union[int, slice]
        :return: Object at index `idx`
        :rtype: Union[Parameter, Descriptor, BaseObj, 'BaseCollection']
        """
        if isinstance(idx, slice):
            start, stop, step = idx.indices(len(self))
            return self.__class__(
                getattr(self, "name"), *[self[i] for i in range(start, stop, step)]
            )
        if str(idx) in self._kwargs.keys():
            return self._kwargs[str(idx)]
        if isinstance(idx, str):
            idx = [index for index, item in enumerate(self) if item.name == idx]
            l = len(idx)
            if l == 0:
                raise IndexError(f"Given index does not exist")
            elif l == 1:
                idx = idx[0]
            else:
                return self.__class__(getattr(self, "name"), *[self[i] for i in idx])
        elif not isinstance(idx, int) or isinstance(idx, bool):
            if isinstance(idx, bool):
                raise TypeError("Boolean indexing is not supported at the moment")
            try:
                if idx > len(self):
                    raise IndexError(f"Given index {idx} is out of bounds")
            except TypeError:
                raise IndexError(
                    "Index must be of type `int`/`slice` or an item name (`str`)"
                )
        keys = list(self._kwargs.keys())
        return self._kwargs[keys[idx]]

    def __setitem__(self, key: int, value: Union[B, V]) -> None:
        """
        Set an item via it's index.

        :param key: Index in self.
        :type key: int
        :param value: Value which index key should be set to.
        :type value: Any
        """
        if isinstance(value, Number):  # noqa: S3827
            item = self.__getitem__(key)
            item.value = value
        elif issubclass(type(value), BasedBase) or issubclass(type(value), Descriptor):
            update_key = list(self._kwargs.keys())
            values = list(self._kwargs.values())
            old_item = values[key]
            # Update the internal dict
            update_dict = {update_key[key]: value}
            self._kwargs.update(update_dict)
            # ADD EDGE
            self._borg.map.add_edge(self, value)
            self._borg.map.reset_type(value, "created_internal")
            value.interface = self.interface
            # REMOVE EDGE
            self._borg.map.prune_vertex_from_edge(self, old_item)
        else:
            raise NotImplementedError(
                "At the moment only numerical values or easyCore objects can be set."
            )

    def __delitem__(self, key: int) -> None:
        """
        Try to delete  an idem by key.

        :param key:
        :type key:
        :return:
        :rtype:
        """
        keys = list(self._kwargs.keys())
        item = self._kwargs[keys[key]]
        self._borg.map.prune_vertex_from_edge(self, item)
        del self._kwargs[keys[key]]

    def __len__(self) -> int:
        """
        Get the number of items in this collection

        :return: Number of items in this collection.
        :rtype: int
        """
        return len(self._kwargs.keys())

    def as_dict(self, skip: Optional[List[str]] = None) -> Dict[str, Any]:
        """
        Convert ones self into a serialized form.

        :return: dictionary of ones self
        :rtype: dict
        """
        d = super(BaseCollection, self).as_dict(skip=skip)
        data = []
        dd = {}
        for key in d.keys():
            if key == "@id":
                continue
            if isinstance(d[key], dict):
                data.append(d[key])
            else:
                dd[key] = d[key]
        dd["data"] = data
        # Attach the id. This might be useful in connected applications.
        # Note that it is converted to int and then str because javascript....
        dd["@id"] = d["@id"]
        return dd

    @property
    def data(self) -> Tuple:
        return tuple(self._kwargs.values())

    @classmethod
    def from_dict(cls, input_dict: Dict[str, Any]) -> "BaseCollection":
        """
        De-serialise the data and try to recreate the object.

        :param input_dict: serialised dictionary of an object. Usually generated from `obj.as_dict()`
        :type input_dict: dict
        :return: Class constructed from the input_dict
        """

        d = input_dict.copy()
        if len(d["data"]) > 0:
            for idx, item in enumerate(d["data"]):
                d[item["@id"]] = item
        del d["data"]
        return super(BaseCollection, cls).from_dict(d)

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__} `{getattr(self, 'name')}` of length {len(self)}"
        )

    def sort(self, mapping: Callable[[Union[B, V]], Any], reverse: bool = False) -> None:
        """
        Sort the collection according to the given mapping.

        :param mapping: mapping function to sort the collection. i.e. lambda parameter: parameter.raw_value
        :type mapping: Callable
        :param reverse: Reverse the sorting.
        :type reverse: bool
        """
        i = list(self._kwargs.items())
        i.sort(key=lambda x: mapping(x[1]), reverse=reverse)
        self._kwargs.reorder(**{k[0]: k[1] for k in i})



#  SPDX-FileCopyrightText: 2022 easyCore contributors  <core@easyscience.software>
#  SPDX-License-Identifier: BSD-3-Clause
#  © 2021-2022 Contributors to the easyCore project <https://github.com/easyScience/easyCore>

__author__ = "github.com/wardsimon"
__version__ = "0.1.0"

import collections.abc
import functools
import warnings

from time import time

from easyCore import borg


class memoized:
    """
    Decorator. Caches a function's return value each time it is called.
    If called later with the same arguments, the cached value is returned
    (not reevaluated).
    """

    def __init__(self, func):
        self.func = func
        self.cache = {}

    def __call__(self, *args):
        if not isinstance(args, collections.abc.Hashable):
            # uncacheable. a list, for instance.
            # better to not cache than blow up.
            return self.func(*args)
        if args in self.cache:
            return self.cache[args]
        value = self.func(*args)
        self.cache[args] = value
        return value

    def __repr__(self) -> str:
        """Return the function's docstring."""
        return self.func.__doc__

    def __get__(self, obj, objtype):
        """Support instance methods."""
        return functools.partial(self.__call__, obj)


def counted(func):
    """
    Counts how many times a function has been called and adds a `func.calls` to it's properties
    :param func: Function to be counted
    :return: Results from function call
    """

    @functools.wraps(func)
    def wrapped(*args, **kwargs):
        wrapped.n_calls += 1
        return func(*args, **kwargs)

    wrapped.n_calls = 0
    return wrapped


def time_it(func):
    """
    Times a function and reports the time either to the class' log or the base logger
    :param func: function to be timed
    :return: callable function with timer
    """
    name = func.__module__ + "." + func.__name__
    time_logger = borg.log.getLogger("timer." + name)

    @functools.wraps(func)
    def _time_it(*args, **kwargs):
        start = int(round(time() * 1000))
        try:
            return func(*args, **kwargs)
        finally:
            end_ = int(round(time() * 1000)) - start
            time_logger.debug(
                f"\033[1;34;49mExecution time: {end_ if end_ > 0 else 0} ms\033[0m"
            )

    return _time_it


def deprecated(func):
    """
    This is a decorator which can be used to mark functions
    as deprecated. It will result in a warning being emitted
    when the function is used.
    """

    @functools.wraps(func)
    def new_func(*args, **kwargs):
        warnings.warn_explicit(
            "Call to deprecated function {}.".format(func.__name__),
            category=DeprecationWarning,
            filename=func.__code__.co_filename,
            lineno=func.__code__.co_firstlineno + 1,
        )
        return func(*args, **kwargs)
    return new_func



#  SPDX-FileCopyrightText: 2022 easyCore contributors  <core@easyscience.software>
#  SPDX-License-Identifier: BSD-3-Clause
#  © 2021-2022 Contributors to the easyCore project <https://github.com/easyScience/easyCore>

__author__ = 'github.com/wardsimon'
__version__ = '0.1.0'

import abc
from collections import deque, UserDict
from typing import Union, Any, NoReturn, Callable, TypeVar

from easyCore import borg


class UndoCommand(metaclass=abc.ABCMeta):
    """
    The Command interface pattern
    """

    def __init__(self, obj) -> None:
        self._obj = obj
        self._text = None

    @abc.abstractmethod
    def undo(self) -> NoReturn:
        """
        Undo implementation which should be overwritten
        """

    @abc.abstractmethod
    def redo(self) -> NoReturn:
        """
        Redo implementation which should be overwritten
        """

    @property
    def text(self) -> str:
        return self._text

    @text.setter
    def text(self, text: str) -> NoReturn:
        self._text = text


T_ = TypeVar('T_', bound=UndoCommand)


def dict_stack_deco(func: Callable) -> Callable:
    def inner(obj, *args, **kwargs):
        # Only do the work to a NotarizedDict.
        if hasattr(obj, '_stack_enabled') and obj._stack_enabled:
            if not kwargs:
                borg.stack.push(DictStack(obj, *args))
            else:
                borg.stack.push(DictStackReCreate(obj, **kwargs))
        else:
            func(obj, *args, **kwargs)
    return inner


class NotarizedDict(UserDict):
    """
    A simple dict drop in for easyCore group classes. This is used as it wraps the get/set methods
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._borg = borg
        self._stack_enabled = False

    @classmethod
    def _classname(cls):
        # This method just returns the name of the class
        return cls.__name__

    @dict_stack_deco
    def __setitem__(self, key, value):
        super(NotarizedDict, self).__setitem__(key, value)

    @dict_stack_deco
    def __delitem__(self, key):
        super(NotarizedDict, self).__delitem__(key)

    def __repr__(self):
        return f"{self._classname()}({self.data})"

    @dict_stack_deco
    def reorder(self, **kwargs):
        self.data = kwargs.copy()


class CommandHolder:
    """
    A holder for one or more commands which are added to the stack
    """

    def __init__(self, text: str = None):
        self._commands = deque()
        self._text = text
        self.__index = 0

    def append(self, command: T_):

        self._commands.appendleft(command)

    def pop(self):
        return self._commands.popleft()

    def __iter__(self) -> T_:
        while self.__index < len(self):
            index = self.__index
            self.__index += 1
            yield self._commands[index]
        self.__index = 0

    def __len__(self) -> int:
        return len(self._commands)

    @property
    def is_macro(self) -> bool:
        return len(self) > 1

    @property
    def current(self) -> T_:
        return self._commands[0]

    @property
    def text(self) -> str:
        text = ''
        if self._commands:
            text = self._commands[-1].text
        if self._text is not None:
            text = self._text
        return text

    @text.setter
    def text(self, text: str):
        self._text = text


class UndoStack:
    """
    Implement a version of QUndoStack without the QT
    """

    def __init__(self, max_history: Union[int, type(None)] = None):
        self._history = deque(maxlen=max_history)
        self._future = deque(maxlen=max_history)
        self._macro_running = False
        self._command_running = False
        self._max_history = max_history
        self._enabled = False

    @property
    def enabled(self) -> bool:
        return self._enabled

    @enabled.setter
    def enabled(self, state: bool):
        if self.enabled and self._macro_running:
            self.endMacro()
        self._enabled = state

    def force_state(self, state: bool):
        self._enabled = state

    @property
    def history(self) -> deque:
        return self._history

    @property
    def future(self) -> deque:
        return self._future

    def push(self, command: T_) -> NoReturn:
        """
        Add a command to the history stack
        """
        # If we're not enabled, then what are we doing!
        if not self.enabled or self._command_running:
            # Do the command and leave.
            command.redo()
            return
        # If there's a macro add the command to the command holder
        if self._macro_running:
            self.history[0].append(command)
        else:
            # Else create the command holder and add it to the stack
            com = CommandHolder()
            com.append(command)
            self.history.appendleft(com)
        # Actually do the command
        command.redo()
        # Reset the future
        self._future = deque(maxlen=self._max_history)

    def pop(self) -> T_:
        """
        !! WARNING - TO BE USED WITH EXTREME CAUTION !!
        !! THIS IS PROBABLY NOT THE FN YOU'RE LOOKING FOR, IT CAN BREAK A LOT OF STUFF !!
        Sometimes you really don't want the last command. Remove it from the stack

        :return: None
        :rtype: None
        """
        pop_it = self._history.popleft()
        popped = pop_it.pop()
        if len(pop_it) > 0:
            self.history.appendleft(pop_it)
        return popped

    def clear(self) -> NoReturn:
        """
        Remove any commands on the stack and reset the state
        """
        self._history = deque(maxlen=self._max_history)
        self._future = deque(maxlen=self._max_history)
        self._macro_running = False

    def undo(self) -> NoReturn:
        """
        Undo the last change to the stack
        """
        if self.canUndo():
            # Move the command from the past to the future
            this_command_stack = self._history.popleft()
            self._future.appendleft(this_command_stack)

            # Execute all undo commands
            for command in this_command_stack:
                try:
                    self._command_running = True
                    command.undo()
                except Exception as e:
                    print(e)
                finally:
                    self._command_running = False

    def redo(self) -> NoReturn:
        """
        Redo the last `undo` command on the stack
        """
        if self.canRedo():

            # Move from the future to the past
            this_command_stack = self._future.popleft()
            self._history.appendleft(this_command_stack)
            # Need to go from right to left
            this_command_stack = list(this_command_stack)
            this_command_stack.reverse()
            for command in this_command_stack:
                try:
                    self._command_running = True
                    command.redo()
                except Exception as e:
                    print(e)
                finally:
                    self._command_running = False

    def beginMacro(self, text: str) -> NoReturn:
        """
        Start a bulk update i.e. multiple commands under one undo/redo command
        """
        if self._macro_running:
            raise AssertionError("Cannot start a macro when one is already running")
        com = CommandHolder(text)
        self.history.appendleft(com)
        self._macro_running = True

    def endMacro(self) -> NoReturn:
        """
        End a bulk update i.e. multiple commands under one undo/redo command
        """
        if not self._macro_running:
            raise AssertionError("Cannot end a macro when one is not running")
        self._macro_running = False

    def canUndo(self) -> bool:
        """
        Can the last command be undone?
        """
        return len(self._history) > 0 and not self._macro_running

    def canRedo(self) -> bool:
        """
        Can we redo a command?
        """
        return len(self._future) > 0 and not self._macro_running

    def redoText(self) -> str:
        """
        Text associated with a redo item.
        """
        text = ''
        if self.canRedo():
            text = self.future[0].text
        return text

    def undoText(self) -> str:
        """
        Text associated with a undo item.
        """
        text = ''
        if self.canUndo():
            text = self.history[0].text
        return text


class PropertyStack(UndoCommand):
    """
    Stack operator for when a property setter is wrapped.
    """

    def __init__(self, parent, func: Callable, old_value: Any, new_value: Any, text: str = None):
        # self.setText("Setting {} to {}".format(func.__name__, new_value))
        super().__init__(self)
        self._parent = parent
        self._old_value = old_value
        self._new_value = new_value
        self._set_func = func
        self.text = f'{parent} value changed from {old_value} to {new_value}'
        if text is not None:
            self.text = text

    def undo(self) -> NoReturn:
        self._set_func(self._parent, self._old_value)

    def redo(self) -> NoReturn:
        self._set_func(self._parent, self._new_value)


class FunctionStack(UndoCommand):
    def __init__(self, parent, set_func: Callable, unset_func: Callable, text: str = None):
        super().__init__(self)
        self._parent = parent
        self._old_fn = set_func
        self._new_fn = unset_func
        self.text = f'{parent} called {set_func}'
        if text is not None:
            self.text = text

    def undo(self):
        self._new_fn()

    def redo(self):
        self._old_fn()


class DictStack(UndoCommand):

    def __init__(self, in_dict: NotarizedDict, *args):
        super().__init__(self)
        self._parent = in_dict

        self._deletion = False
        self._creation = False

        self._key = None
        self._index = None
        self._old_value = None
        self._new_value = None
        self.text = ''

        if len(args) == 1:
            # We are deleting
            self._deletion = True
            self._index = list(self._parent.keys()).index(args[0])
            self._old_value = self._parent[args[0]]
            self._key = args[0]
            self.text = f'Deleting {args[0]} from {self._parent}'
        elif len(args) == 2:
            # We are either creating or setting
            self._key = args[0]
            self._new_value = args[1]
            if self._key in self._parent.keys():
                # We are modifying
                self._old_value = self._parent[self._key]
                self.text = f'Setting {self._parent}[{self._key}] from {self._old_value} to {self._new_value}'
            else:
                self._creation = True
                self.text = f'Creating {self._parent}[{self._key}] with value {self._new_value}'
        else:
            raise ValueError

    def undo(self) -> NoReturn:
        if self._creation:
            # Now we delete
            self._parent.data.__delitem__(self._key)
        else:
            # Now we create/change value
            if self._index is None:
                self._parent.data.__setitem__(self._key, self._old_value)
            else:
                # This deals with placing an item in a place
                keys = list(self._parent.keys())
                values = list(self._parent.values())
                keys.insert(self._index, self._key)
                values.insert(self._index, self._old_value)
                self._parent.reorder(**{k: v for k, v in zip(keys, values)})

    def redo(self) -> NoReturn:
        if self._deletion:
            # Now we delete
            self._parent.data.__delitem__(self._key)
        else:
            self._parent.data.__setitem__(self._key, self._new_value)


class DictStackReCreate(UndoCommand):

    def __init__(self, in_dict: NotarizedDict, **kwargs):
        super().__init__(self)
        self._parent = in_dict
        self._old_value = in_dict.data.copy()
        self._new_value = kwargs
        self.text = 'Updating dictionary'

    def undo(self) -> NoReturn:
        self._parent.data = self._old_value

    def redo(self) -> NoReturn:
        self._parent.data = self._new_value


def property_stack_deco(arg: Union[str, Callable], begin_macro: bool = False) -> Callable:
    """
    Decorate a `property` setter with undo/redo functionality
    This decorator can be used as:

    @property_stack_deco
    def func()
    ....

    or

    @property_stack_deco("This is the undo/redo text)
    def func()
    ....

    In the latter case the argument is a string which might be evaluated.
    The possible markups for this string are;

    `obj` - The thing being operated on
    `func` - The function being called
    `name` - The name of the function being called.
    `old_value` - The pre-set value
    `new_value` - The post-set value

    An example would be `Function {name}: Set from {old_value} to {new_value}`

    """
    if isinstance(arg, Callable):
        func = arg
        name = func.__name__

        def wrapper(obj, *args) -> NoReturn:
            old_value = getattr(obj, name)
            new_value = args[0]

            if new_value == old_value:
                return

            if borg.debug:
                print(f"I'm {obj} and have been set from {old_value} to {new_value}!")

            borg.stack.push(PropertyStack(obj, func, old_value, new_value))

        setattr(wrapper, 'func', func)
    else:
        txt = arg

        def wrapper(func: Callable) -> Callable:

            name = func.__name__

            def inner_wrapper(obj, *args) -> NoReturn:

                if begin_macro:
                    borg.stack.beginMacro(txt)

                old_value = getattr(obj, name)
                new_value = args[0]

                if new_value == old_value:
                    return

                if borg.debug:
                    print(f"I'm {obj} and have been set from {old_value} to {new_value}!")
                borg.stack.push(PropertyStack(obj, func, old_value, new_value, text=txt.format(**locals())))

            setattr(inner_wrapper, 'func', func)
            return inner_wrapper
    return wrapper



#  SPDX-FileCopyrightText: 2022 easyCore contributors  <core@easyscience.software>
#  SPDX-License-Identifier: BSD-3-Clause
#  © 2021-2022 Contributors to the easyCore project <https://github.com/easyScience/easyCore>

__author__ = 'github.com/wardsimon'
__version__ = '0.1.0'

import warnings
imported = -1
try:
    from easyCore.Fitting.lmfit import lmfit  # noqa: F401, E402
    imported += 1
except ImportError:
    # TODO make this a proper message (use logging?)
    warnings.warn('lmfit has not been installed.', ImportWarning, stacklevel=2)
try:
    from easyCore.Fitting.bumps import bumps  # noqa: F401, E402
    imported += 1
except ImportError:
    # TODO make this a proper message (use logging?)
    warnings.warn('bumps has not been installed.', ImportWarning, stacklevel=2)
try:
    from easyCore.Fitting.DFO_LS import DFO  # noqa: F401, E402
    imported += 1
except ImportError:
    # TODO make this a proper message (use logging?)
    warnings.warn('dfo-ls has not been installed.', ImportWarning, stacklevel=2)

from easyCore.Fitting.fitting_template import FittingTemplate  # noqa: E402

engines: list = FittingTemplate._engines



from __future__ import annotations

#  SPDX-FileCopyrightText: 2022 easyCore contributors  <core@easyscience.software>
#  SPDX-License-Identifier: BSD-3-Clause
#  © 2021-2022 Contributors to the easyCore project <https://github.com/easyScience/easyCore>

__author__ = "github.com/wardsimon"
__version__ = "0.1.0"

from abc import ABCMeta, abstractmethod
from typing import TypeVar, List, NamedTuple, Callable, TYPE_CHECKING, Optional, Type

from easyCore import np


_C = TypeVar("_C", bound=ABCMeta)
_M = TypeVar("_M")
if TYPE_CHECKING:
    from easyCore.Fitting.Fitting import Fitter


class InterfaceFactoryTemplate:
    """
    This class allows for the creation and transference of interfaces.
    """

    def __init__(self, interface_list: List[_C], *args, **kwargs):
        self._interfaces: List[_C] = interface_list
        self._current_interface: _C
        self.__interface_obj: _M = None
        self.create(*args, **kwargs)

    def create(self, *args, **kwargs):
        """
        Create an interface to a calculator from those initialized. Interfaces can be selected
        by `interface_name` where `interface_name` is one of `obj.available_interfaces`. This
        interface can now be accessed by obj().

        :param interface_name: name of interface to be created
        :type interface_name: str
        :return: None
        :rtype: noneType
        """
        if kwargs.get("interface_name", None) is None:
            if len(self._interfaces) > 0:
                # Fallback name
                interface_name = self.return_name(self._interfaces[0])
            else:
                raise NotImplementedError
        else:
            interface_name = kwargs.pop("interface_name")
        interfaces = self.available_interfaces
        if interface_name in interfaces:
            self._current_interface = self._interfaces[interfaces.index(interface_name)]
        self.__interface_obj = self._current_interface(*args, **kwargs)

    def switch(self, new_interface: str, fitter: Optional[Type[Fitter]] = None):
        """
        Changes the current interface to a new interface. The current interface is destroyed and
        all MSONable parameters carried over to the new interface. i.e. pick up where you left off.

        :param new_interface: name of new interface to be created
        :type new_interface: str
        :param fitter: Fitting interface which contains the fitting object which may have bindings which will be updated.
        :type fitter: easyCore.Fitting.Fitting.Fitter
        :return: None
        :rtype: noneType
        """
        interfaces = self.available_interfaces
        if new_interface in interfaces:
            self._current_interface = self._interfaces[interfaces.index(new_interface)]
            self.__interface_obj = self._current_interface()
        else:
            raise AttributeError("The user supplied interface is not valid.")
        if fitter is not None:
            if hasattr(fitter, "_fit_object"):
                obj = getattr(fitter, "_fit_object")
                try:
                    if hasattr(obj, "update_bindings"):
                        obj.update_bindings()
                except Exception as e:
                    print(f"Unable to auto generate bindings.\n{e}")
            elif hasattr(fitter, "generate_bindings"):
                try:
                    fitter.generate_bindings()
                except Exception as e:
                    print(f"Unable to auto generate bindings.\n{e}")

    @property
    def available_interfaces(self) -> List[str]:
        """
        Return all available interfaces.

        :return: List of available interface names
        :rtype: List[str]
        """
        return [self.return_name(this_interface) for this_interface in self._interfaces]

    @property
    def current_interface(self) -> _C:
        """
        Returns the constructor for the currently selected interface.

        :return: Interface constructor
        :rtype: InterfaceTemplate
        """
        return self._current_interface

    @property
    def current_interface_name(self) -> str:
        """
        Returns the constructor name for the currently selected interface.

        :return: Interface constructor name
        :rtype: str
        """
        return self.return_name(self._current_interface)

    @property
    def fit_func(
        self,
    ) -> Callable:  # , x_array: np.ndarray, *args, **kwargs) -> np.ndarray:
        """
        Pass through to the underlying interfaces fitting function.

        :param x_array: points to be calculated at
        :type x_array: np.ndarray
        :param args: positional arguments for the fitting function
        :type args: Any
        :param kwargs: key/value pair arguments for the fitting function.
        :type kwargs: Any
        :return: points calculated at positional values `x`
        :rtype: np.ndarray
        #"""

        def __fit_func(*args, **kwargs):
            return self.__interface_obj.fit_func(*args, **kwargs)
        return __fit_func

    def generate_bindings(self, model, *args, ifun=None, **kwargs):
        """
        Automatically bind a `Parameter` to the corresponding interface.
        :param name: parameter name
        :type name: str
        :return: binding property
        :rtype: property
        """
        class_links = self.__interface_obj.create(model)
        props = model._get_linkable_attributes()
        props_names = [prop.name for prop in props]
        for item in class_links:
            for item_key in item.name_conversion.keys():
                if item_key not in props_names:
                    continue
                idx = props_names.index(item_key)
                prop = props[idx]
                prop._callback = item.make_prop(item_key)
                prop._callback.fset(prop.raw_value)

    def __call__(self, *args, **kwargs) -> _M:
        return self.__interface_obj

    def __reduce__(self):
        return (
            self.__state_restore__,
            (
                self.__class__,
                self.current_interface_name,
            ),
        )

    @staticmethod
    def __state_restore__(cls, interface_str):
        obj = cls()
        if interface_str in obj.available_interfaces:
            obj.switch(interface_str)
        return obj

    @staticmethod
    def return_name(this_interface) -> str:
        """
        Return an interfaces name
        """
        interface_name = this_interface.__name__
        if hasattr(this_interface, "name"):
            interface_name = getattr(this_interface, "name")
        return interface_name


class ItemContainer(NamedTuple):
    link_name: str
    name_conversion: dict
    getter_fn: Callable
    setter_fn: Callable

    def make_prop(self, parameter_name) -> property:
        return property(
            fget=self.__make_getter(parameter_name),
            fset=self.__make_setter(parameter_name),
        )

    def convert_key(self, lookup_key: str) -> str:
        key = self.name_conversion.get(lookup_key, None)
        return key

    def __make_getter(self, get_name: str) -> Callable:
        def get_value():
            inner_key = self.name_conversion.get(get_name, None)
            return self.getter_fn(self.link_name, inner_key)

        return get_value

    def __make_setter(self, get_name: str) -> Callable:
        def set_value(value):
            inner_key = self.name_conversion.get(get_name, None)
            self.setter_fn(self.link_name, **{inner_key: value})

        return set_value

iF = TypeVar("iF", bound=InterfaceFactoryTemplate)



#  SPDX-FileCopyrightText: 2022 easyCore contributors  <core@easyscience.software>
#  SPDX-License-Identifier: BSD-3-Clause
#  © 2021-2022 Contributors to the easyCore project <https://github.com/easyScience/easyCore>

__author__ = 'github.com/wardsimon'
__version__ = '0.1.0'

import weakref
import inspect
import sys
from typing import Tuple, List
from collections.abc import Callable
from abc import ABCMeta, abstractmethod
from functools import wraps
from types import MethodType

from easyCore.Utils.classUtils import singleton
from collections.abc import Mapping


@singleton
class Store:
    __log = []
    __var_ident = 'var_'
    __ret_ident = 'ret_'

    def __init__(self):
        self.log = self.__log  # TODO Async problem?
        self.var_ident = self.__var_ident
        self.ret_ident = self.__ret_ident

    @staticmethod
    def get_defaults() -> dict:
        return {
            'log':         Store.__log,
            'create_list': Store.__create_list,
            'unique_args': Store.__unique_args,
            'unique_rets': Store.__unique_rets,
            'var_ident':   Store.__var_ident,
            'ret_ident':   Store.__ret_ident
        }

    def append_log(self, log_entry: str):
        self.log.append(log_entry)


class ScriptManager:
    def __init__(self, enabled=True):
        self._store = Store()
        self._enabled = enabled

    @property
    def enabled(self) -> bool:
        return self._enabled

    @enabled.setter
    def enabled(self, value: bool):
        self._enabled = value

    def history(self) -> List[str]:
        return self._store.log

    def reset_history(self):
        defaults = Store.get_defaults()
        for key, item in defaults.items():
            setattr(self._store, key, item)

    def append_log(self, log_entry: str):
        self._store.log.append(log_entry)


class Hugger(metaclass=ABCMeta):

    def __init__(self):
        self._store = Store()

    @property
    def log(self) -> list:
        return self._store.log

    @abstractmethod
    def patch(self):
        pass

    @abstractmethod
    def restore(self):
        pass


class PatcherFactory(Hugger, metaclass=ABCMeta):

    def __init__(self):
        super().__init__()

    @staticmethod
    def is_mutable(arg) -> bool:
        ret = True
        if isinstance(arg, (int, float, complex, str, tuple, frozenset, bytes, property)):
            ret = False
        return ret

    @staticmethod
    def _caller_name(skip: int = 2):
        """Get a name of a caller in the format module.class.method
           `skip` specifies how many levels of stack to skip while getting caller
           name. skip=1 means "who calls me", skip=2 "who calls my caller" etc.
           An empty string is returned if skipped levels exceed stack height
           https://gist.github.com/techtonik/2151727#gistcomment-2333747
        """

        def stack_(frame):
            framelist = []
            while frame:
                framelist.append(frame)
                frame = frame.f_back
            return framelist

        stack = stack_(sys._getframe(1))
        start = 0 + skip
        if len(stack) < start + 1:
            return ''
        parentframe = stack[start]

        name = []
        module = inspect.getmodule(parentframe)
        # `modname` can be None when frame is executed directly in console
        # TODO(techtonik): consider using __main__
        if module:
            name.append(module.__name__)
        # detect classname
        if 'self' in parentframe.f_locals:
            # I don't know any way to detect call from the object method
            # XXX: there seems to be no way to detect static method call - it will
            #      be just a function call
            name.append(parentframe.f_locals['self'].__class__.__name__)
        codename = parentframe.f_code.co_name
        if codename != '<module>':  # top level usually
            name.append(codename)  # function or a method
        del parentframe
        return ".".join(name)

    def _append_args(self, *args, **kwargs):

        def check(res):
            return id(res) not in self._store.unique_rets and \
                   id(res) not in self._store.create_list and \
                   id(res) not in self._store.unique_args

        for arg in args:
            if self.is_mutable(arg) and check(arg):
                self._store.unique_args.append(id(arg))
        for item in kwargs.values():
            if self.is_mutable(item) and check(item):
                self._store.unique_args.append(id(item))

    def _append_create(self, obj):
        this_id = id(obj)
        if this_id not in self._store.create_list:
            self._store.create_list.append(this_id)

    def _append_result(self, result) -> int:
        ret = 0

        def check(res):
            return id(res) not in self._store.unique_rets and \
                   id(res) not in self._store.create_list and \
                   id(res) not in self._store.unique_args

        if isinstance(result, type(None)):
            return ret
        elif isinstance(result, tuple):
            for res in result:
                # if self.is_mutable(res) and check(res):
                if check(res):
                    self._store.unique_rets.append(id(res))
            ret = len(result)
        else:
            # if self.is_mutable(result) and check(result):
            if check(result):
                self._store.unique_rets.append(id(result))
            ret = 1
        return ret

    def _append_log(self, log_entry: str):
        self._store.log.append(log_entry)

    def __options(self, item) -> Tuple[int, dict]:
        this_id = id(item)
        option = {
            'create_list': self._store.create_list,
            'return_list': self._store.unique_rets,
            'input_list':  self._store.unique_args
        }
        return this_id, option

    def _get_position(self, query: str, item) -> int:
        this_id, option = self.__options(item)
        in_list = self._in_list(query, item)
        index = None
        if in_list:
            index = option.get(query).index(this_id)
        return index

    def _in_list(self, query: str, item) -> bool:
        this_id, option = self.__options(item)
        return this_id in option.get(query, [])

    @staticmethod
    def _get_class_that_defined_method(method_in) -> classmethod:
        if inspect.ismethod(method_in):
            for cls in inspect.getmro(method_in.__self__.__class__):
                if cls.__dict__.get(method_in.__name__) is method_in:
                    return cls
            method_in = method_in.__func__  # fallback to __qualname__ parsing
        if inspect.isfunction(method_in):
            class_name = method_in.__qualname__.split('.<locals>', 1)[0].rsplit('.', 1)[0]
            try:
                cls = getattr(inspect.getmodule(method_in), class_name)
            except AttributeError:
                cls = method_in.__globals__.get(class_name)
            if isinstance(cls, type):
                return cls



#  SPDX-FileCopyrightText: 2022 easyCore contributors  <core@easyscience.software>
#  SPDX-License-Identifier: BSD-3-Clause
#  © 2021-2022 Contributors to the easyCore project <https://github.com/easyScience/easyCore>

__author__ = "github.com/wardsimon"
__version__ = "0.1.0"

import inspect
from typing import List, Optional

from easyCore.Fitting.fitting_template import (
    Union,
    Callable,
    FittingTemplate,
    np,
    FitResults,
    NameConverter,
    FitError,
)

# Import lmfit specific objects
from lmfit import Parameter as lmParameter, Parameters as lmParameters, Model as lmModel
from lmfit.model import ModelResult


class lmfit(FittingTemplate):  # noqa: S101
    """
    This is a wrapper to lmfit: https://lmfit.github.io/
    It allows for the lmfit fitting engine to use parameters declared in an `easyCore.Objects.Base.BaseObj`.
    """

    property_type = lmParameter
    name = "lmfit"

    def __init__(self, obj, fit_function: Callable):
        """
        Initialize the fitting engine with a `BaseObj` and an arbitrary fitting function.

        :param obj: Object containing elements of the `Parameter` class
        :type obj: BaseObj
        :param fit_function: function that when called returns y values. 'x' must be the first
                            and only positional argument. Additional values can be supplied by
                            keyword/value pairs
        :type fit_function: Callable
        """
        super().__init__(obj, fit_function)

    def make_model(self, pars: Optional[lmParameters] = None) -> lmModel:
        """
        Generate a lmfit model from the supplied `fit_function` and parameters in the base object.

        :return: Callable lmfit model
        :rtype: lmModel
        """
        # Generate the fitting function
        fit_func = self._generate_fit_function()
        if pars is None:
            pars = self._cached_pars
        # Create the model
        model = lmModel(
            fit_func,
            independent_vars=["x"],
            param_names=["p" + str(key) for key in pars.keys()],
        )
        # Assign values from the `Parameter` to the model
        for name, item in pars.items():
            if isinstance(item, lmParameter):
                value = item.value
            else:
                value = item.raw_value
            model.set_param_hint(
                "p" + str(name), value=value, min=item.min, max=item.max
            )

        # Cache the model for later reference
        self._cached_model = model
        return model

    def _generate_fit_function(self) -> Callable:
        """
        Using the user supplied `fit_function`, wrap it in such a way we can update `Parameter` on
        iterations.

        :return: a fit function which is compatible with lmfit models
        :rtype: Callable
        """
        # Original fit function
        func = self._original_fit_function
        # Get a list of `Parameters`
        self._cached_pars = {}
        for parameter in self._object.get_fit_parameters():
            self._cached_pars[NameConverter().get_key(parameter)] = parameter

        # Make a new fit function
        def fit_function(x: np.ndarray, **kwargs):
            """
            Wrapped fit function which now has a lmfit compatible form.

            :param x: array of data points to be calculated
            :type x: np.ndarray
            :param kwargs: key word arguments
            :return: points calculated at `x`
            :rtype: np.ndarray
            """
            # Update the `Parameter` values and the callback if needed
            # TODO THIS IS NOT THREAD SAFE :-(
            for name, value in kwargs.items():
                par_name = int(name[1:])
                if par_name in self._cached_pars.keys():
                    # This will take in to account constraints
                    self._cached_pars[par_name].value = value
                    # Since we are calling the parameter fset will be called.
            # TODO Pre processing here
            for constraint in self.fit_constraints():
                constraint()
            return_data = func(x)
            # TODO Loading or manipulating data here
            return return_data

        # Fake the function signature.
        # This is done as lmfit wants the function to be in the form:
        # f = (x, a=1, b=2)...
        # Where we need to be generic. Note that this won't hold for much outside of this scope.
        params = [
            inspect.Parameter(
                "x", inspect.Parameter.POSITIONAL_OR_KEYWORD, annotation=inspect._empty
            ),
            *[
                inspect.Parameter(
                    "p" + str(name),
                    inspect.Parameter.POSITIONAL_OR_KEYWORD,
                    annotation=inspect._empty,
                    default=parameter.raw_value,
                )
                for name, parameter in self._cached_pars.items()
            ],
        ]
        # Sign the function
        fit_function.__signature__ = inspect.Signature(params)
        self._fit_function = fit_function
        return fit_function

    def fit(
        self,
        x: np.ndarray,
        y: np.ndarray,
        weights: Optional[np.ndarray] = None,
        model: Optional[lmModel] = None,
        parameters: Optional[lmParameters] = None,
        method: Optional[str] = None,
        minimizer_kwargs: Optional[dict] = None,
        engine_kwargs: Optional[dict] = None,
        **kwargs,
    ) -> FitResults:
        """
        Perform a fit using the lmfit engine.

        :param method:
        :type method:
        :param x: points to be calculated at
        :type x: np.ndarray
        :param y: measured points
        :type y: np.ndarray
        :param weights: Weights for supplied measured points
        :type weights: np.ndarray
        :param model: Optional Model which is being fitted to
        :type model: lmModel
        :param parameters: Optional parameters for the fit
        :type parameters: lmParameters
        :param minimizer_kwargs: Arguments to be passed directly to the minimizer
        :type minimizer_kwargs: dict
        :param kwargs: Additional arguments for the fitting function.
        :return: Fit results
        :rtype: ModelResult
        """
        default_method = {}
        if method is not None and method in self.available_methods():
            default_method["method"] = method

        if weights is None:
            weights = np.sqrt(np.abs(y))

        if engine_kwargs is None:
            engine_kwargs = {}

        if minimizer_kwargs is None:
            minimizer_kwargs = {}
        else:
            minimizer_kwargs = {"fit_kws": minimizer_kwargs}
        minimizer_kwargs.update(engine_kwargs)

        # Why do we do this? Because a fitting template has to have borg instantiated outside pre-runtime
        from easyCore import borg

        try:
            borg.stack.beginMacro("Fitting routine")
            if model is None:
                model = self.make_model()

            model_results = model.fit(
                y, x=x, weights=weights, **default_method, **minimizer_kwargs, **kwargs
            )
            self._set_parameter_fit_result(model_results)
            results = self._gen_fit_results(model_results)
        except Exception as e:
            raise FitError(e)
        finally:
            borg.stack.endMacro()
        return results

    def convert_to_pars_obj(
        self, par_list: Optional[List] = None
    ) -> lmParameters:
        """
        Create an lmfit compatible container with the `Parameters` converted from the base object.

        :param par_list: If only a single/selection of parameter is required. Specify as a list
        :type par_list: List[str]
        :return: lmfit Parameters compatible object
        :rtype: lmParameters
        """
        if par_list is None:
            # Assume that we have a BaseObj for which we can obtain a list
            par_list = self._object.get_fit_parameters()
        pars_obj = lmParameters().add_many(
            [self.__class__.convert_to_par_object(obj) for obj in par_list]
        )
        return pars_obj

    @staticmethod
    def convert_to_par_object(obj) -> lmParameter:
        """
        Convert an `easyCore.Objects.Base.Parameter` object to a lmfit Parameter object.

        :return: lmfit Parameter compatible object.
        :rtype: lmParameter
        """
        return lmParameter(
            "p" + str(NameConverter().get_key(obj)),
            value=obj.raw_value,
            vary=not obj.fixed,
            min=obj.min,
            max=obj.max,
            expr=None,
            brute_step=None,
        )

    def _set_parameter_fit_result(self, fit_result: ModelResult):
        """
        Update parameters to their final values and assign a std error to them.

        :param fit_result: Fit object which contains info on the fit
        :return: None
        :rtype: noneType
        """
        pars = self._cached_pars
        for name in pars.keys():
            pars[name].value = fit_result.params["p" + str(name)].value
            if fit_result.errorbars:
                pars[name].error = fit_result.params["p" + str(name)].stderr
            else:
                pars[name].error = 0.0

    def _gen_fit_results(self, fit_results: ModelResult, **kwargs) -> FitResults:
        """
        Convert fit results into the unified `FitResults` format.
        See https://github.com/lmfit/lmfit-py/blob/480072b9f7834b31ff2ca66277a5ad31246843a4/lmfit/model.py#L1272

        :param fit_result: Fit object which contains info on the fit
        :return: fit results container
        :rtype: FitResults
        """
        results = FitResults()
        for name, value in kwargs.items():
            if getattr(results, name, False):
                setattr(results, name, value)

        # We need to unify return codes......
        results.success = fit_results.success
        results.y_obs = fit_results.data
        results.residual = fit_results.residual
        results.x = fit_results.userkws["x"]
        results.p = fit_results.values
        results.p0 = fit_results.init_values
        results.goodness_of_fit = fit_results.chisqr
        results.y_calc = fit_results.best_fit

        results.fitting_engine = self.__class__
        results.fit_args = None

        results.engine_result = fit_results
        # results.check_sanity()
        return results

    def available_methods(self) -> List[str]:
        return [
            "least_squares",
            "leastsq",
            "differential_evolution",
            "basinhopping",
            "ampgo",
            "nelder",
            "lbfgsb",
            "powell",
            "cg",
            "newton",
            "cobyla",
            "bfgs",
        ]



#  SPDX-FileCopyrightText: 2022 easyCore contributors  <core@easyscience.software>
#  SPDX-License-Identifier: BSD-3-Clause
#  © 2021-2022 Contributors to the easyCore project <https://github.com/easyScience/easyCore>

from __future__ import annotations

__author__ = "github.com/wardsimon"
__version__ = "0.1.0"

from typing import List, Tuple, TYPE_CHECKING, TypeVar, Union

from easyCore import borg
from easyCore.Utils.Hugger.Property import LoggedProperty

if TYPE_CHECKING:
    from easyCore.Utils.typing import B, BV

def addLoggedProp(inst: BV, name: str, *args, **kwargs) -> None:
    cls = type(inst)
    annotations = getattr(cls, "__annotations__", False)
    if not hasattr(cls, "__perinstance"):
        cls = type(cls.__name__, (cls,), {"__module__": inst.__module__})
        cls.__perinstance = True
        if annotations:
            cls.__annotations__ = annotations
        inst.__old_class__ = inst.__class__
        inst.__class__ = cls
    setattr(cls, name, LoggedProperty(*args, **kwargs))


def addProp(inst: BV, name: str, *args, **kwargs) -> None:
    cls = type(inst)
    annotations = getattr(cls, "__annotations__", False)
    if not hasattr(cls, "__perinstance"):
        cls = type(cls.__name__, (cls,), {"__module__": __name__})
        cls.__perinstance = True
        if annotations:
            cls.__annotations__ = annotations
        inst.__old_class__ = inst.__class__
        inst.__class__ = cls

    setattr(cls, name, property(*args, **kwargs))


def removeProp(inst: BV, name: str) ->None:
    cls = type(inst)
    if not hasattr(cls, "__perinstance"):
        cls = type(cls.__name__, (cls,), {"__module__": __name__})
        cls.__perinstance = True
        inst.__old_class__ = inst.__class__
        inst.__class__ = cls
    delattr(cls, name)


def generatePath(model_obj: B, skip_first: bool = False) -> Tuple[List[int], List[str]]:
    pars = model_obj.get_parameters()
    start_idx = 0 + int(skip_first)
    ids = []
    names = []
    model_id = borg.map.convert_id(model_obj)
    for par in pars:
        elem = borg.map.convert_id(par)
        route = borg.map.reverse_route(elem, model_id)
        objs = [getattr(borg.map.get_item_by_key(r), "name") for r in route]
        objs.reverse()
        names.append(".".join(objs[start_idx:]))
        ids.append(elem.int)
    return ids, names



from __future__ import annotations

__author__ = "github.com/wardsimon"
__version__ = "0.1.0"

#  SPDX-FileCopyrightText: 2022 easyCore contributors  <core@easyscience.software>
#  SPDX-License-Identifier: BSD-3-Clause
#  © 2021-2022 Contributors to the easyCore project <https://github.com/easyScience/easyCore>

import numbers

from typing import (
    List,
    Union,
    Iterable,
    Dict,
    Optional,
    TypeVar,
    TYPE_CHECKING,
    Callable,
)

from easyCore import borg
from easyCore.Utils.json import MSONable
from easyCore.Utils.classTools import addLoggedProp
from .Variable import Parameter, Descriptor

if TYPE_CHECKING:
    from easyCore.Utils.typing import C, V, iF


class BasedBase(MSONable):
    __slots__ = ["_name", "_borg", "user_data", "_kwargs"]

    def __init__(self, name: str, interface: Optional[iF] = None):
        self._borg = borg
        self._borg.map.add_vertex(self, obj_type="created")
        self.interface = interface
        self.user_data: dict = {}
        self._name: str = name

    def __reduce__(self):
        """
        Make the class picklable.
        Due to the nature of the dynamic class definitions special measures need to be taken.

        :return: Tuple consisting of how to make the object
        :rtype: tuple
        """
        state = self.as_dict()
        cls = getattr(self, "__old_class__", self.__class__)
        return cls.from_dict, (state,)

    @property
    def name(self) -> str:
        """
        Get the common name of the object.

        :return: Common name of the object
        """
        return self._name

    @name.setter
    def name(self, new_name: str):
        """
        Set a new common name for the object.

        :param new_name: New name for the object
        :return: None
        """
        self._name = new_name

    @property
    def interface(self) -> iF:
        """
        Get the current interface of the object
        """
        return self._interface

    @interface.setter
    def interface(self, new_interface: iF):
        """
        Set the current interface to the object and generate bindings if possible. iF.e.
        ```
        def __init__(self, bar, interface=None, **kwargs):
            super().__init__(self, **kwargs)
            self.foo = bar
            self.interface = interface # As final step after initialization to set correct bindings.
        ```
        """
        self._interface = new_interface
        if new_interface is not None:
            self.generate_bindings()

    def generate_bindings(self):
        """
        Generate or re-generate bindings to an interface (if exists)

        :raises: AttributeError
        """
        if self.interface is None:
            raise AttributeError(
                "Interface error for generating bindings. `interface` has to be set."
            )
        interfaceable_children = [
            key
            for key in self._borg.map.get_edges(self)
            if issubclass(type(self._borg.map.get_item_by_key(key)), BasedBase)
        ]
        for child_key in interfaceable_children:
            child = self._borg.map.get_item_by_key(child_key)
            child.interface = self.interface
        self.interface.generate_bindings(self)

    def switch_interface(self, new_interface_name: str):
        """
        Switch or create a new interface.
        """
        if self.interface is None:
            raise AttributeError(
                "Interface error for generating bindings. `interface` has to be set."
            )
        self.interface.switch(new_interface_name)
        self.generate_bindings()

    @property
    def constraints(self) -> List[C]:
        pars = self.get_parameters()
        constraints = []
        for par in pars:
            con: Dict[str, C] = par.user_constraints
            for key in con.keys():
                constraints.append(con[key])
        return constraints

    def as_dict(
        self, skip: Optional[List[str]] = None
    ) -> Dict[str, Union[str, bool, numbers.Number]]:
        """
        Convert ones self into a serialized form.

        :return: dictionary of ones self
        """
        if skip is None:
            skip = []
        d = MSONable.as_dict(self, skip=skip)
        for key, item in d.items():
            if hasattr(item, "as_dict"):
                d[key] = item.as_dict(skip=skip)
        # Attach the id. This might be useful in connected applications.
        # Note that it is converted to int and then str because javascript....
        d["@id"] = str(self._borg.map.convert_id(self).int)
        return d

    def get_parameters(self) -> List[Parameter]:
        """
        Get all parameter objects as a list.

        :return: List of `Parameter` objects.
        """
        par_list = []
        for key, item in self._kwargs.items():
            if hasattr(item, "get_parameters"):
                par_list = [*par_list, *item.get_parameters()]
            elif isinstance(item, Parameter):
                par_list.append(item)
        return par_list

    def _get_linkable_attributes(self) -> List[V]:
        """
        Get all objects which can be linked against as a list.

        :return: List of `Descriptor`/`Parameter` objects.
        """
        item_list = []
        for key, item in self._kwargs.items():
            if hasattr(item, "_get_linkable_attributes"):
                item_list = [*item_list, *item._get_linkable_attributes()]
            elif issubclass(type(item), Descriptor):
                item_list.append(item)
        return item_list

    def get_fit_parameters(self) -> List[Parameter]:
        """
        Get all objects which can be fitted (and are not fixed) as a list.

        :return: List of `Parameter` objects which can be used in fitting.
        """
        fit_list = []
        for key, item in self._kwargs.items():
            if hasattr(item, "get_fit_parameters"):
                fit_list = [*fit_list, *item.get_fit_parameters()]
            elif isinstance(item, Parameter) and item.enabled and not item.fixed:
                fit_list.append(item)
        return fit_list

    def __dir__(self) -> Iterable[str]:
        """
        This creates auto-completion and helps out in iPython notebooks.

        :return: list of function and parameter names for auto-completion
        """
        new_class_objs = list(k for k in dir(self.__class__) if not k.startswith("_"))
        return sorted(new_class_objs)


if TYPE_CHECKING:
    B = TypeVar("B", bound=BasedBase)
    BV = TypeVar("BV", bound=Union[BasedBase, Descriptor])


class BaseObj(BasedBase):
    """
    This is the base class for which all higher level classes are built off of.
    NOTE: This object is serializable only if parameters are supplied as:
    `BaseObj(a=value, b=value)`. For `Parameter` or `Descriptor` objects we can
    cheat with `BaseObj(*[Descriptor(...), Parameter(...), ...])`.
    """

    def __init__(
        self,
        name: str,
        *args: Optional[BV],
        **kwargs: Optional[BV],
    ):
        """
        Set up the base class.

        :param name: Name of this object
        :param args: Any arguments?
        :param kwargs: Fields which this class should contain
        """
        super(BaseObj, self).__init__(name)
        # If Parameter or Descriptor is given as arguments...
        for arg in args:
            if issubclass(type(arg), (BaseObj, Descriptor)):
                kwargs[getattr(arg, "name")] = arg
        # Set kwargs, also useful for serialization
        known_keys = self.__dict__.keys()
        self._kwargs = kwargs
        for key in kwargs.keys():
            if key in known_keys:
                raise AttributeError
            if issubclass(
                type(kwargs[key]), (BasedBase, Descriptor)
            ) or "BaseCollection" in [c.__name__ for c in type(kwargs[key]).__bases__]:
                self._borg.map.add_edge(self, kwargs[key])
                self._borg.map.reset_type(kwargs[key], "created_internal")
            addLoggedProp(
                self,
                key,
                self.__getter(key),
                self.__setter(key),
                get_id=key,
                my_self=self,
                test_class=BaseObj,
            )

    def _add_component(self, key: str, component: BV) -> None:
        """
        Dynamically add a component to the class. This is an internal method, though can be called remotely.
        The recommended alternative is to use typing, i.e.

        class Foo(Bar):
            def __init__(self, foo: Parameter, bar: Parameter):
                super(Foo, self).__init__(bar=bar)
                self._add_component("foo", foo)

        Goes to:
         class Foo(Bar):
            foo: ClassVar[Parameter]
            def __init__(self, foo: Parameter, bar: Parameter):
                super(Foo, self).__init__(bar=bar)
                self.foo = foo

        :param key: Name of component to be added
        :param component: Component to be added
        :return: None
        """
        self._kwargs[key] = component
        self._borg.map.add_edge(self, component)
        self._borg.map.reset_type(component, "created_internal")
        addLoggedProp(
            self,
            key,
            self.__getter(key),
            self.__setter(key),
            get_id=key,
            my_self=self,
            test_class=BaseObj,
        )

    def __setattr__(self, key: str, value: BV) -> None:
        # Assume that the annotation is a ClassVar
        old_obj = None
        if (
            hasattr(self.__class__, "__annotations__")
            and key in self.__class__.__annotations__
            and hasattr(self.__class__.__annotations__[key], "__args__")
            and issubclass(
                getattr(value, "__old_class__", value.__class__),
                self.__class__.__annotations__[key].__args__,
            )
        ):
            if issubclass(type(getattr(self, key, None)), (BasedBase, Descriptor)):
                old_obj = self.__getattribute__(key)
                self._borg.map.prune_vertex_from_edge(self, old_obj)
            self._add_component(key, value)
        else:
            if hasattr(self, key) and issubclass(type(value), (BasedBase, Descriptor)):
                old_obj = self.__getattribute__(key)
                self._borg.map.prune_vertex_from_edge(self, old_obj)
                self._borg.map.add_edge(self, value)
        super(BaseObj, self).__setattr__(key, value)
        # Update the interface bindings if something changed (BasedBase and Descriptor)
        if old_obj is not None:
            old_interface = getattr(self, "interface", None)
            if old_interface is not None:
                self.generate_bindings()

    def __repr__(self) -> str:
        return f"{self.__class__.__name__} `{getattr(self, 'name')}`"

    @staticmethod
    def __getter(key: str) -> Callable[[BV], BV]:
        def getter(obj: BV) -> BV:
            return obj._kwargs[key]

        return getter

    @staticmethod
    def __setter(key: str) -> Callable[[BV], None]:
        def setter(obj: BV, value: float) -> None:
            if issubclass(obj._kwargs[key].__class__, Descriptor) and not issubclass(
                value.__class__, Descriptor
            ):
                obj._kwargs[key].value = value
            else:
                obj._kwargs[key] = value

        return setter

    # @staticmethod
    # def __setter(key: str) -> Callable[[Union[B, V]], None]:
    #     def setter(obj: Union[V, B], value: float) -> None:
    #         if issubclass(obj._kwargs[key].__class__, Descriptor):
    #             if issubclass(obj._kwargs[key].__class__, Descriptor):
    #                 obj._kwargs[key] = value
    #             else:
    #                 obj._kwargs[key].value = value
    #         else:
    #             obj._kwargs[key] = value
    #
    #     return setter



#  SPDX-FileCopyrightText: 2022 easyCore contributors  <core@easyscience.software>
#  SPDX-License-Identifier: BSD-3-Clause
#  © 2022 Contributors to the easyCore project <https://github.com/easyScience/easyCore>

from __future__ import annotations

__author__ = "github.com/wardsimon"
__version__ = "0.0.1"

import inspect
import weakref
from copy import deepcopy
from typing import Iterable, MutableSequence, TYPE_CHECKING, TypeVar

from easyCore import borg
from easyCore.Fitting.Constraints import ObjConstraint


def raise_(ex):
    raise ex


def _remover(a_obj_id: str, v_obj_id: str):
    try:
        # Try to get parent object (might be deleted)
        a_obj = borg.map.get_item_by_key(int(a_obj_id))
    except ValueError:
        return
    if a_obj._constraints["virtual"].get(v_obj_id, False):
        del a_obj._constraints["virtual"][v_obj_id]


def realizer(obj):
    if getattr(obj, "_is_virtual", False):
        klass = getattr(obj, "__non_virtual_class__")
        import easyCore.Objects.Variable as ec_var
        args = []
        if klass in ec_var.__dict__.values():  # is_variable check
            kwargs = obj.to_data_dict()
            return klass(**kwargs)
        else:
            kwargs = {name: realizer(item) for name, item in obj._kwargs.items()}
            if isinstance(klass, Iterable) or issubclass(klass, MutableSequence):
                for key, value in inspect.signature(klass).parameters.items():
                    if value.kind in [inspect.Parameter.POSITIONAL_ONLY,
                                      inspect.Parameter.POSITIONAL_OR_KEYWORD]:
                        args.append(getattr(obj, key))
            return klass(*args, **kwargs)
    else:
        return obj


def component_realizer(obj, component, recursive=True):
    import easyCore.Objects.Variable as ec_var
    done_mapping = True
    if not isinstance(obj, Iterable) or not issubclass(obj.__class__, MutableSequence):
        old_component = obj._kwargs[component]
        new_components = realizer(obj._kwargs[component])
        if hasattr(new_components,'enabled'):
            new_components.enabled = True
    else:
        old_component = obj[component]
        new_components = realizer(obj[component])
        idx = obj.index(old_component)
        del obj[component]
        obj.insert(idx, new_components)
        done_mapping = False
    if not recursive:
        for key in iter(component):
            if isinstance(key, str):
                value = component._kwargs[key]
            else:
                value = key
                key = value._borg.map.convert_id_to_key(value)
            if getattr(value, '__old_class__', value.__class__) in ec_var.__dict__.values():
                continue
            component._borg.map.prune_vertex_from_edge(component, component._kwargs[key])
            component._borg.map.add_edge(component, old_component._kwargs[key])
            component._kwargs[key] = old_component._kwargs[key]
            done_mapping = False
    if done_mapping:
        obj._borg.map.prune_vertex_from_edge(obj, old_component)
        obj._borg.map.add_edge(obj, new_components)
        obj._kwargs[component] = new_components
    return obj


def virtualizer(obj):
    """
    Convert a real `easyCore` object to a virtual object.
    This means that the object returned is an exact copy which is unsettable, unchangeable
    and linked to the parent object.
    The object can be realized and returned as a copy via the `realizer` function. If you need a
    component realized in place then `relalize_component` should be called.

    :param obj:
    :type obj:
    :return:
    :rtype:
    """
    # First  check if we're already a virtual object
    if getattr(obj, "_is_virtual", False):
        new_obj = deepcopy(obj)
        old_obj = obj._borg.map.get_item_by_key(obj._derived_from)
        constraint = ObjConstraint(new_obj, "", old_obj)
        constraint.external = True
        old_obj._constraints["virtual"][str(obj._borg.map.convert_id(new_obj).int)] = constraint
        new_obj._constraints["builtin"] = dict()
        # setattr(new_obj, "__previous_set", getattr(olobj, "__previous_set", None))
        weakref.finalize(
            new_obj,
            _remover,
            str(borg.map.convert_id(old_obj).int),
            str(borg.map.convert_id(new_obj).int),
        )
        return new_obj

    # The supplied class
    klass = getattr(obj, "__old_class__", obj.__class__)
    virtual_options = {
        "_is_virtual":           True,
        "is_virtual":            property(fget=lambda self: self._is_virtual),
        "_derived_from":         property(fget=lambda self: self._borg.map.convert_id(obj).int),
        "__non_virtual_class__": klass,
        "realize":               realizer,
        "relalize_component":    component_realizer,
    }

    import easyCore.Objects.Variable as ec_var

    if klass in ec_var.__dict__.values():  # is_variable check
        virtual_options["fixed"] = property(
            fget=lambda self: self._fixed,
            fset=lambda self, value: raise_(
                AttributeError("Virtual parameters cannot be fixed")
            ),
        )
    # Generate a new class
    cls = type("Virtual" + klass.__name__, (klass,), virtual_options)
    # Determine what to do next.
    args = []
    # If `obj` is a parameter or descriptor etc, then simple mods.
    if hasattr(obj, "_constructor"):
        # All Variables are based on the Descriptor.
        d = obj.as_dict()
        if hasattr(d, "fixed"):
            d["fixed"] = True
        v_p = cls.from_dict(d)
        v_p._enabled = False
        constraint = ObjConstraint(v_p, "", obj)
        constraint.external = True
        obj._constraints["virtual"][str(cls._borg.map.convert_id(v_p).int)] = constraint
        v_p._constraints["builtin"] = dict()
        setattr(v_p, "__previous_set", getattr(obj, "__previous_set", None))
        weakref.finalize(
            v_p,
            _remover,
            str(borg.map.convert_id(obj).int),
            str(borg.map.convert_id(v_p).int),
        )
    else:
        # In this case, we need to be recursive.
        kwargs = {name: virtualizer(item) for name, item in obj._kwargs.items()}
        if isinstance(klass, Iterable) or issubclass(klass, MutableSequence):
            for key, value in inspect.signature(cls).parameters.items():
                if value.kind in [inspect.Parameter.POSITIONAL_ONLY,
                                  inspect.Parameter.POSITIONAL_OR_KEYWORD]:
                    args.append(getattr(obj, key))
        v_p = cls(*args, **kwargs)
    return v_p



#  SPDX-FileCopyrightText: 2022 easyCore contributors  <core@easyscience.software>
#  SPDX-License-Identifier: BSD-3-Clause
#  © 2021-2022 Contributors to the easyCore project <https://github.com/easyScience/easyCore>

__author__ = "github.com/wardsimon"
__version__ = "0.1.0"

import weakref
import sys
from typing import List, Union
from weakref import WeakKeyDictionary
from collections import defaultdict
from uuid import uuid4, UUID


class _EntryList(list):
    def __init__(self, *args, my_type=None, **kwargs):
        super(_EntryList, self).__init__(*args, **kwargs)
        self.__known_types = {"argument", "created", "created_internal", "returned"}
        self.finalizer = None
        self._type = []
        if my_type in self.__known_types:
            self._type.append(my_type)

    def __repr__(self) -> str:
        s = "Graph entry of type: "
        if self._type:
            s += ", ".join(self._type)
        else:
            s += "Undefined"
        s += ". With"
        if self.finalizer is None:
            s += "out"
        s += "a finalizer."
        return s

    def __delitem__(self, key):
        super(_EntryList, self).__delitem__(key)

    def remove_type(self, old_type: str):
        if old_type in self.__known_types and old_type in self._type:
            self._type.remove(old_type)

    def reset_type(self, default_type: str = None):
        self._type = []
        self.type = default_type

    @property
    def type(self) -> List[str]:
        return self._type

    @type.setter
    def type(self, value: str):
        if value in self.__known_types and value not in self._type:
            self._type.append(value)

    @property
    def is_argument(self) -> bool:
        return "argument" in self._type

    @property
    def is_created(self) -> bool:
        return "created" in self._type

    @property
    def is_created_internal(self) -> bool:
        return "created_internal" in self._type

    @property
    def is_returned(self) -> bool:
        return "returned" in self._type


class UniqueIdMap(WeakKeyDictionary):
    def __init__(self, this_dict: dict = None):
        super().__init__(self)
        # replace data with a defaultdict to generate uuids
        self.data = defaultdict(uuid4)
        if this_dict is not None:
            self.update(this_dict)


uniqueidmap = UniqueIdMap()


class Graph:
    def __init__(self):
        self._store = weakref.WeakValueDictionary()
        self.__graph_dict = {}

    def vertices(self) -> List[int]:
        """returns the vertices of a graph"""
        return list(self._store.keys())

    def edges(self):
        """returns the edges of a graph"""
        return self.__generate_edges()

    @property
    def argument_objs(self) -> List[int]:
        return self._nested_get("argument")

    @property
    def created_objs(self) -> List[int]:
        return self._nested_get("created")

    @property
    def created_internal(self) -> List[int]:
        return self._nested_get("created_internal")

    @property
    def returned_objs(self) -> List[int]:
        return self._nested_get("returned")

    def get_item_by_key(self, item_id: int) -> object:
        if item_id in self._store.keys():
            return self._store[item_id]
        raise ValueError

    def is_known(self, vertex: object) -> bool:
        return self.convert_id(vertex).int in self._store.keys()

    def find_type(self, vertex: object) -> List[str]:
        if self.is_known(vertex):
            oid = self.convert_id(vertex)
            return self.__graph_dict[oid].type

    def reset_type(self, obj, default_type: str):
        if self.convert_id(obj).int in self.__graph_dict.keys():
            self.__graph_dict[self.convert_id(obj).int].reset_type(default_type)

    def change_type(self, obj, new_type: str):
        if self.convert_id(obj).int in self.__graph_dict.keys():
            self.__graph_dict[self.convert_id(obj).int].type = new_type

    def add_vertex(self, obj: object, obj_type: str = None):
        oid = self.convert_id(obj).int
        self._store[oid] = obj
        self.__graph_dict[oid] = _EntryList()  # Enhanced list of keys
        self.__graph_dict[oid].finalizer = weakref.finalize(
            self._store[oid], self.prune, oid
        )
        self.__graph_dict[oid].type = obj_type

    def add_edge(self, start_obj: object, end_obj: object):
        vertex1 = self.convert_id(start_obj).int
        vertex2 = self.convert_id(end_obj).int
        if vertex1 in self.__graph_dict.keys():
            self.__graph_dict[vertex1].append(vertex2)
        else:
            raise AttributeError

    def get_edges(self, start_obj) -> List[str]:
        vertex1 = self.convert_id(start_obj).int
        if vertex1 in self.__graph_dict.keys():
            return list(self.__graph_dict[vertex1])
        else:
            raise AttributeError

    def __generate_edges(self) -> list:
        """A static method generating the edges of the
        graph "graph". Edges are represented as sets
        with one (a loop back to the vertex) or two
        vertices
        """
        edges = []
        for vertex in self.__graph_dict:
            for neighbour in self.__graph_dict[vertex]:
                if {neighbour, vertex} not in edges:
                    edges.append({vertex, neighbour})
        return edges

    def prune_vertex_from_edge(self, parent_obj, child_obj):
        vertex1 = self.convert_id(parent_obj).int
        if child_obj is None:
            return
        vertex2 = self.convert_id(child_obj).int

        if (
            vertex1 in self.__graph_dict.keys()
            and vertex2 in self.__graph_dict[vertex1]
        ):
            del self.__graph_dict[vertex1][self.__graph_dict[vertex1].index(vertex2)]

    def prune(self, key: int):
        if key in self.__graph_dict.keys():
            del self.__graph_dict[key]

    def find_isolated_vertices(self) -> list:
        """returns a list of isolated vertices."""
        graph = self.__graph_dict
        isolated = []
        for vertex in graph:
            print(isolated, vertex)
            if not graph[vertex]:
                isolated += [vertex]
        return isolated

    def find_path(self, start_obj, end_obj, path=[]) -> list:
        """find a path from start_vertex to end_vertex
        in graph"""

        try:
            start_vertex = self.convert_id(start_obj).int
            end_vertex = self.convert_id(end_obj).int
        except TypeError:
            start_vertex = start_obj
            end_vertex = end_obj

        graph = self.__graph_dict
        path = path + [start_vertex]
        if start_vertex == end_vertex:
            return path
        if start_vertex not in graph:
            return []
        for vertex in graph[start_vertex]:
            if vertex not in path:
                extended_path = self.find_path(vertex, end_vertex, path)
                if extended_path:
                    return extended_path
        return []

    def find_all_paths(self, start_obj, end_obj, path=[]) -> list:
        """find all paths from start_vertex to
        end_vertex in graph"""

        start_vertex = self.convert_id(start_obj).int
        end_vertex = self.convert_id(end_obj).int

        graph = self.__graph_dict
        path = path + [start_vertex]
        if start_vertex == end_vertex:
            return [path]
        if start_vertex not in graph:
            return []
        paths = []
        for vertex in graph[start_vertex]:
            if vertex not in path:
                extended_paths = self.find_all_paths(vertex, end_vertex, path)
                for p in extended_paths:
                    paths.append(p)
        return paths

    def reverse_route(self, end_obj, start_obj=None) -> List:
        """
        In this case we have an object and want to know the connections to get to another in reverse.
        We might not know the start_object. In which case we follow the shortest path to a base vertex.
        :param end_obj:
        :type end_obj:
        :param start_obj:
        :type start_obj:
        :return:
        :rtype:
        """
        end_vertex = self.convert_id(end_obj).int

        path_length = sys.maxsize
        optimum_path = []
        if start_obj is None:
            # We now have to find where to begin.....
            for possible_start, vertices in self.__graph_dict.items():
                if end_vertex in vertices:
                    temp_path = self.find_path(possible_start, end_vertex)
                    if len(temp_path) < path_length:
                        path_length = len(temp_path)
                        optimum_path = temp_path
        else:
            optimum_path = self.find_path(start_obj, end_obj)
        optimum_path.reverse()
        return optimum_path

    def is_connected(self, vertices_encountered=None, start_vertex=None) -> bool:
        """determines if the graph is connected"""
        if vertices_encountered is None:
            vertices_encountered = set()
        graph = self.__graph_dict
        vertices = list(graph.keys())
        if not start_vertex:
            # chose a vertex from graph as a starting point
            start_vertex = vertices[0]
        vertices_encountered.add(start_vertex)
        if len(vertices_encountered) != len(vertices):
            for vertex in graph[start_vertex]:
                if vertex not in vertices_encountered and self.is_connected(
                    vertices_encountered, vertex
                ):
                    return True
        else:
            return True
        return False

    def _nested_get(self, obj_type: str) -> List[int]:
        """Access a nested object in root by key sequence."""
        extracted_list = []
        for key, item in self.__graph_dict.items():
            if obj_type in item.type:
                extracted_list.append(key)
        return extracted_list

    @staticmethod
    def convert_id(input_value) -> UUID:
        """Sometimes we're dopy and"""
        if not validate_id(input_value):
            input_value = unique_id(input_value)
        return input_value

    @staticmethod
    def convert_id_to_key(input_value: Union[object, UUID]) -> int:
        """Sometimes we're dopy and"""
        if not validate_id(input_value):
            input_value: UUID = unique_id(input_value)
        return input_value.int

    def __repr__(self) -> str:
        return f"Graph object of {len(self._store)} vertices."


def unique_id(obj) -> UUID:
    """Produce a unique integer id for the object.

    Object must me *hashable*. Id is a UUID and should be unique
    across Python invocations.

    """
    return uniqueidmap[obj]


def validate_id(potential_id) -> bool:
    test = True
    try:
        if isinstance(potential_id, UUID):
            UUID(str(potential_id), version=4)
        else:
            UUID(potential_id, version=4)
    except (ValueError, AttributeError):
        test = False
    return test



#  SPDX-FileCopyrightText: 2022 easyCore contributors  <core@easyscience.software>
#  SPDX-License-Identifier: BSD-3-Clause
#  © 2021-2022 Contributors to the easyCore project <https://github.com/easyScience/easyCore>

__author__ = 'github.com/wardsimon'
__version__ = '0.0.1'
